import numpy as np
import pytest
from numpy.testing import assert_allclose
from ..MeanFunction import MeanFunction, MeanBase, MeanSum, MeanProduct, FixedMean, ConstantMean
from ..MeanFunction import LinearMean, Coefficient, PolynomialMean, MeanComposite, MeanPower
from ..MeanFunction import fixed_f, fixed_inputderiv, one, const_f, const_deriv

@pytest.fixture
def mf():
    return MeanBase()

@pytest.fixture
def x():
    return np.array([[1., 2., 3.], [4., 5., 6.]])

@pytest.fixture
def params():
    return np.array([1., 2., 4.])

@pytest.fixture
def zeroparams():
    return np.zeros(0)

@pytest.fixture
def oneparams():
    return np.ones(1)

@pytest.fixture
def dx():
    return 1.e-6

def test_MeanBase(mf, x, params):
    "test composition of mean functions"

    mf2 = mf + mf

    assert isinstance(mf2, MeanSum)

    mf3 = 5. + mf

    assert isinstance(mf3, MeanSum)
    assert isinstance(mf3.f1, ConstantMean)

    mf4 = mf + 3.

    assert isinstance(mf4, MeanSum)
    assert isinstance(mf4.f2, ConstantMean)

    mf5 = mf*mf

    assert isinstance(mf5, MeanProduct)

    mf6 = 3.*mf

    assert isinstance(mf6, MeanProduct)
    assert isinstance(mf6.f1, ConstantMean)

    mf7 = mf*3.

    assert isinstance(mf7, MeanProduct)
    assert isinstance(mf7.f2, ConstantMean)

    mf8 = mf**2

    assert isinstance(mf8, MeanPower)
    assert isinstance(mf8.f2, ConstantMean)

    mf9 = mf**Coefficient()

    assert isinstance(mf9, MeanPower)

    mf10 = 2.**Coefficient()

    assert isinstance(mf10, MeanPower)
    assert isinstance(mf10.f1, ConstantMean)

    mf11 = mf(mf)

    assert isinstance(mf11, MeanComposite)

def test_MeanBase_failures(mf, x, params):
    "test situations of MeanBase where an exception should be raised"

    with pytest.raises(NotImplementedError):
        mf.get_n_params(x)

    with pytest.raises(NotImplementedError):
        mf.mean_f(x, params)

    with pytest.raises(NotImplementedError):
        mf.mean_deriv(x, params)

    with pytest.raises(NotImplementedError):
        mf.mean_hessian(x, params)

    with pytest.raises(NotImplementedError):
        mf.mean_inputderiv(x, params)

    with pytest.raises(TypeError):
        "3" + mf

    with pytest.raises(TypeError):
        mf + "3"

    with pytest.raises(TypeError):
        mf*"3"

    with pytest.raises(TypeError):
        "3"*mf

    with pytest.raises(TypeError):
        mf**mf

    with pytest.raises(TypeError):
        2.**mf

    with pytest.raises(TypeError):
        mf**"2"

    with pytest.raises(TypeError):
        "2"**mf

    with pytest.raises(TypeError):
        mf(3.)

def test_MeanFunction():
    "test the function to create a mean function from a formula"

    mf = MeanFunction(None)

    assert isinstance(mf, ConstantMean)
    assert_allclose(mf.mean_f(np.ones((2,3)), np.zeros(0)), np.zeros((2)))

    mf2 = MeanFunction("x[0]", use_patsy=True)

    assert isinstance(mf2, MeanSum)
    assert_allclose(mf2.mean_f(np.array([[1., 2., 3.], [4., 5., 6.]]), np.array([1., 2.])),
                    np.array([3., 9.]))

    mf3 = MeanFunction("a + b*c", {"c":0}, use_patsy=False)

    assert isinstance(mf3, MeanSum)
    assert_allclose(mf3.mean_f(np.array([[1., 2., 3.], [4., 5., 6.]]), np.array([1., 2.])),
                    np.array([3., 9.]))

    with pytest.raises(ValueError):
        MeanFunction(1)

def test_fixed_functions(x):
    "test utility functions for creating fixed means"

    assert_allclose(fixed_f(np.ones((2,2)), 0, np.exp), np.exp(np.ones(2)))

    with pytest.raises(AssertionError):
        fixed_f(np.ones(3), 0, np.exp)
    with pytest.raises(AssertionError):
        fixed_f(np.ones((2,2)), -1, np.exp)
    with pytest.raises(AssertionError):
        fixed_f(np.ones((2,2)), 0, 1)
    with pytest.raises(IndexError):
        fixed_f(np.ones((2,2)), 3, np.exp)

    inputderiv_exp = np.zeros((2,2))
    inputderiv_exp[0] = np.exp(1.)

    assert_allclose(fixed_inputderiv(np.ones((2,2)), 0, np.exp), inputderiv_exp)

    with pytest.raises(AssertionError):
        fixed_inputderiv(np.ones(3), 0, np.exp)
    with pytest.raises(AssertionError):
        fixed_inputderiv(np.ones((2,2)), -1, np.exp)
    with pytest.raises(AssertionError):
        fixed_inputderiv(np.ones((2,2)), 0, 1)
    with pytest.raises(IndexError):
        fixed_inputderiv(np.ones((2,2)), 3, np.exp)

    assert_allclose(one(x), np.ones(x.shape))
    assert_allclose(const_f(x, 2.), np.broadcast_to(2., (x.shape[0],)))
    assert_allclose(const_deriv(x), np.zeros((x.shape[1], x.shape[0])))

def test_FixedMean(x, zeroparams):
    "test the FixedMean function"

    f = lambda x: np.ones(x.shape[0])
    deriv = lambda x: np.zeros((x.shape[1], x.shape[0]))
    fixed_mean = FixedMean(f, deriv)

    assert fixed_mean.get_n_params(x) == 0

    assert_allclose(fixed_mean.mean_f(x, zeroparams), np.ones(x.shape[0]))
    assert_allclose(fixed_mean.mean_f(np.array([1., 2., 3.]), zeroparams), np.ones(3))
    assert_allclose(fixed_mean.mean_deriv(x, zeroparams), np.zeros((0, x.shape[0])))
    assert_allclose(fixed_mean.mean_hessian(x, zeroparams), np.zeros((0, 0, x.shape[0])))
    assert_allclose(fixed_mean.mean_inputderiv(x, zeroparams), np.zeros((x.shape[1], x.shape[0])))

    fixed_mean = FixedMean(f)

    with pytest.raises(RuntimeError):
        fixed_mean.mean_inputderiv(x, np.zeros(0))

@pytest.mark.parametrize("val",[0., 1., 2.])
def test_ConstantMean(x, zeroparams, val):
    "test the constant mean function"

    constant_mean = ConstantMean(val)

    assert constant_mean.get_n_params(x) == 0

    assert_allclose(constant_mean.mean_f(x, zeroparams), np.broadcast_to(val, x.shape[0]))
    assert_allclose(constant_mean.mean_deriv(x, zeroparams), np.zeros((0, x.shape[0])))
    assert_allclose(constant_mean.mean_hessian(x, zeroparams), np.zeros((0, 0, x.shape[0])))
    assert_allclose(constant_mean.mean_inputderiv(x, zeroparams), np.zeros((x.shape[1], x.shape[0])))

def test_ConstantMean_failures():
    "test situation where ConstantMean should fail"

    with pytest.raises(TypeError):
        ConstantMean("a")

@pytest.mark.parametrize("index", [0, 1])
def test_LinearMean(x, zeroparams, dx, index):
    "test the linear mean function"

    linear_mean = LinearMean(index)

    assert linear_mean.get_n_params(x) == 0

    inputderiv_expect = np.zeros((x.shape[1], x.shape[0]))
    inputderiv_expect[index] = 1.

    assert_allclose(linear_mean.mean_f(x, zeroparams), x[:,index])
    assert_allclose(linear_mean.mean_deriv(x, zeroparams), np.zeros((0, x.shape[0])))
    assert_allclose(linear_mean.mean_hessian(x, zeroparams), np.zeros((0, 0, x.shape[0])))
    assert_allclose(linear_mean.mean_inputderiv(x, zeroparams), inputderiv_expect)

    D = x.shape[1]

    inputderiv_fd = np.zeros((D, 2))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (linear_mean.mean_f(x, zeroparams) -
                            linear_mean.mean_f(x - dx_array, zeroparams))/dx

    assert_allclose(linear_mean.mean_inputderiv(x, zeroparams), inputderiv_fd)

def test_Coefficient(x, oneparams):
    "test the Coefficient class"

    coeff_mean = Coefficient()

    assert coeff_mean.get_n_params(x) == 1

    assert_allclose(coeff_mean.mean_f(x, oneparams), np.broadcast_to(1., x.shape[0]))
    assert_allclose(coeff_mean.mean_deriv(x, oneparams), np.ones((1, x.shape[0])))
    assert_allclose(coeff_mean.mean_hessian(x, oneparams), np.zeros((1, 1, x.shape[0])))
    assert_allclose(coeff_mean.mean_inputderiv(x, oneparams), np.zeros((x.shape[1], x.shape[0])))

    with pytest.raises(AssertionError):
        coeff_mean._check_inputs(x, np.ones(4))

    with pytest.raises(AssertionError):
        coeff_mean._check_inputs(np.ones((2, 3, 2)), oneparams)

def test_LinearMean_failures(x):
    "test situations where LinearMean should fail"

    mf = LinearMean(3)

    with pytest.raises(IndexError):
        mf.mean_f(x, np.zeros(0))

def test_MeanSum(x, oneparams, dx):
    "test the MeanSum function"

    n_params = 1
    index = 0

    mf = Coefficient() + LinearMean(index)

    assert mf.get_n_params(x) == n_params

    inputderiv_exp = np.zeros((x.shape[1], x.shape[0]))
    inputderiv_exp[index] = 1.

    assert_allclose(mf.mean_f(x, oneparams), 1. + x[:,index])
    assert_allclose(mf.mean_deriv(x, oneparams), np.ones((n_params, x.shape[0])))
    assert_allclose(mf.mean_hessian(x, oneparams), np.zeros((n_params, n_params, x.shape[0])))
    assert_allclose(mf.mean_inputderiv(x, oneparams), inputderiv_exp)

    deriv_fd = np.zeros((n_params, x.shape[0]))
    for i in range(n_params):
        dx_array = np.zeros(n_params)
        dx_array[i] = dx
        deriv_fd[i] = (mf.mean_f(x, oneparams) - mf.mean_f(x, oneparams - dx_array))/dx

    assert_allclose(mf.mean_deriv(x, oneparams), deriv_fd)

    hess_fd = np.zeros((n_params, n_params, x.shape[0]))
    for i in range(n_params):
        for j in range(n_params):
            dx_array = np.zeros(n_params)
            dx_array[j] = dx
            hess_fd[i,j] = (mf.mean_deriv(x, oneparams)[i] -
                            mf.mean_deriv(x, oneparams - dx_array)[i])/dx

    assert_allclose(mf.mean_hessian(x, oneparams), hess_fd)

    D = x.shape[1]

    inputderiv_fd = np.zeros((D, x.shape[0]))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (mf.mean_f(x, oneparams) - mf.mean_f(x - dx_array, oneparams))/dx

    assert_allclose(mf.mean_inputderiv(x, oneparams), inputderiv_fd)

def test_MeanProduct(x, oneparams, dx):
    "test the MeanProduct function"

    n_params = 1
    index = 0

    mf = Coefficient()*LinearMean(index)

    assert mf.get_n_params(x) == n_params

    inputderiv_exp = np.zeros((x.shape[1], x.shape[0]))
    inputderiv_exp[index] = np.broadcast_to(oneparams, (x.shape[0],))

    assert_allclose(mf.mean_f(x, oneparams), 1.*x[:,index])
    assert_allclose(mf.mean_deriv(x, oneparams), np.broadcast_to(x[:,index], (n_params, x.shape[0])))
    assert_allclose(mf.mean_hessian(x, oneparams), np.zeros((n_params, n_params, x.shape[0])))
    assert_allclose(mf.mean_inputderiv(x, oneparams), inputderiv_exp)

    deriv_fd = np.zeros((n_params, x.shape[0]))
    for i in range(n_params):
        dx_array = np.zeros(n_params)
        dx_array[i] = dx
        deriv_fd[i] = (mf.mean_f(x, oneparams) - mf.mean_f(x, oneparams - dx_array))/dx

    assert_allclose(mf.mean_deriv(x, oneparams), deriv_fd)

    hess_fd = np.zeros((n_params, n_params, x.shape[0]))
    for i in range(n_params):
        for j in range(n_params):
            dx_array = np.zeros(n_params)
            dx_array[j] = dx
            hess_fd[i,j] = (mf.mean_deriv(x, oneparams)[i] -
                            mf.mean_deriv(x, oneparams - dx_array)[i])/dx

    assert_allclose(mf.mean_hessian(x, oneparams), hess_fd)

    D = x.shape[1]

    inputderiv_fd = np.zeros((D, x.shape[0]))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (mf.mean_f(x, oneparams) - mf.mean_f(x - dx_array, oneparams))/dx

    assert_allclose(mf.mean_inputderiv(x, oneparams), inputderiv_fd)

def test_MeanSum_MeanProduct_combination(x, params, dx):
    "test a more complicated combination of sums and products"

    n_params = 3
    index1 = 0
    index2 = 1
    index3 = 2

    mf = (Coefficient() + Coefficient()*LinearMean(index1) +
          Coefficient()*LinearMean(index2)*LinearMean(index3))

    assert mf.get_n_params(x) == n_params

    deriv_exp = np.zeros((n_params, x.shape[0]))
    deriv_exp[0] = 1.
    deriv_exp[1] = x[:,index1]
    deriv_exp[2] = x[:,index2]*x[:,index3]

    inputderiv_exp = np.zeros((x.shape[1], x.shape[0]))
    inputderiv_exp[0,:] = params[1]
    inputderiv_exp[1,:] = params[2]*x[:,index3]
    inputderiv_exp[2,:] = params[2]*x[:,index2]

    assert_allclose(mf.mean_f(x, params), params[0] + params[1]*x[:,index1]
                                                    + params[2]*x[:,index2]*x[:,index3])
    assert_allclose(mf.mean_deriv(x, params), deriv_exp)
    assert_allclose(mf.mean_hessian(x, params), np.zeros((n_params, n_params, x.shape[0])))
    assert_allclose(mf.mean_inputderiv(x, params), inputderiv_exp)

    deriv_fd = np.zeros((n_params, x.shape[0]))
    for i in range(n_params):
        dx_array = np.zeros(n_params)
        dx_array[i] = dx
        deriv_fd[i] = (mf.mean_f(x, params) - mf.mean_f(x, params - dx_array))/dx

    assert_allclose(mf.mean_deriv(x, params), deriv_fd)

    hess_fd = np.zeros((n_params, n_params, x.shape[0]))
    for i in range(n_params):
        for j in range(n_params):
            dx_array = np.zeros(n_params)
            dx_array[j] = dx
            hess_fd[i,j] = (mf.mean_deriv(x, params)[i] -
                            mf.mean_deriv(x, params - dx_array)[i])/dx

    assert_allclose(mf.mean_hessian(x, params), hess_fd)

    D = x.shape[1]

    inputderiv_fd = np.zeros((D, x.shape[0]))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (mf.mean_f(x, params) - mf.mean_f(x - dx_array, params))/dx

    assert_allclose(mf.mean_inputderiv(x, params), inputderiv_fd)

def test_MeanPower(x, zeroparams, oneparams, dx):
    "test the power mean function"

    mf = LinearMean(0)**2

    assert mf.get_n_params(x) == 0

    inputderiv_exp = np.zeros((x.shape[1], x.shape[0]))
    inputderiv_exp[0,:] = 2.*x[:,0]

    assert_allclose(mf.mean_f(x, zeroparams), x[:,0]**2)
    assert_allclose(mf.mean_deriv(x, zeroparams), np.zeros((0, x.shape[0])))
    assert_allclose(mf.mean_hessian(x, zeroparams), np.zeros((0, 0, x.shape[0])))
    assert_allclose(mf.mean_inputderiv(x, zeroparams), inputderiv_exp)

    D = x.shape[1]

    inputderiv_fd = np.zeros((D, x.shape[0]))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (mf.mean_f(x, zeroparams) - mf.mean_f(x - dx_array, zeroparams))/dx

    assert_allclose(mf.mean_inputderiv(x, zeroparams), inputderiv_fd, atol=1.e-6, rtol=1.e-6)

    mf2 = LinearMean(0)**1

    inputderiv_exp = np.zeros((x.shape[1], x.shape[0]))
    inputderiv_exp[0,:] = np.ones(x.shape[0])

    assert_allclose(mf2.mean_f(x, zeroparams), x[:,0])
    assert_allclose(mf2.mean_deriv(x, zeroparams), np.zeros((0, x.shape[0])))
    assert_allclose(mf2.mean_hessian(x, zeroparams), np.zeros((0, 0, x.shape[0])))
    assert_allclose(mf2.mean_inputderiv(x, zeroparams), inputderiv_exp)

    inputderiv_fd = np.zeros((D, x.shape[0]))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (mf2.mean_f(x, zeroparams) - mf2.mean_f(x - dx_array, zeroparams))/dx

    assert_allclose(mf2.mean_inputderiv(x, zeroparams), inputderiv_fd, atol=1.e-6, rtol=1.e-6)

    mf3 = LinearMean(0)**0

    assert_allclose(mf3.mean_f(x, zeroparams), np.ones(x.shape[0]))
    assert_allclose(mf3.mean_deriv(x, zeroparams), np.zeros((0, x.shape[0])))
    assert_allclose(mf3.mean_hessian(x, zeroparams), np.zeros((0, 0, x.shape[0])))
    assert_allclose(mf3.mean_inputderiv(x, zeroparams), np.zeros((x.shape[1], x.shape[0])))

    inputderiv_fd = np.zeros((D, x.shape[0]))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (mf3.mean_f(x, zeroparams) - mf3.mean_f(x - dx_array, zeroparams))/dx

    assert_allclose(mf3.mean_inputderiv(x, zeroparams), inputderiv_fd, atol=1.e-6, rtol=1.e-6)

    params = np.array([2.1])
    n_params = 1

    mf4 = LinearMean(0)**Coefficient()

    assert mf4.get_n_params(x) == n_params

    deriv_exp_0 = np.zeros((n_params, x.shape[0]))
    deriv_exp_0[0] = np.log(x[:,0])*x[:,0]**0
    deriv_exp_1 = np.zeros((n_params, x.shape[0]))
    deriv_exp_1[0] = np.log(x[:,0])*x[:,0]**oneparams[0]
    deriv_exp_2 = np.zeros((n_params, x.shape[0]))
    deriv_exp_2[0] = np.log(x[:,0])*x[:,0]**params[0]
    hess_exp_0 = np.zeros((n_params, n_params, x.shape[0]))
    hess_exp_0[0, 0] = (np.log(x[:,0])**2)*x[:,0]**0
    hess_exp_1 = np.zeros((n_params, n_params, x.shape[0]))
    hess_exp_1[0, 0] = (np.log(x[:,0])**2)*x[:,0]**oneparams[0]
    hess_exp_2 = np.zeros((n_params, n_params, x.shape[0]))
    hess_exp_2[0, 0] = (np.log(x[:,0])**2)*x[:,0]**params[0]
    inputderiv_exp_1 = np.zeros((x.shape[1], x.shape[0]))
    inputderiv_exp_1[0] = oneparams[0]*x[:,0]**(oneparams[0] - 1.)
    inputderiv_exp_2 = np.zeros((x.shape[1], x.shape[0]))
    inputderiv_exp_2[0] = params[0]*x[:,0]**(params[0] - 1.)

    assert_allclose(mf4.mean_f(x, oneparams), x[:,0]**oneparams[0])
    assert_allclose(mf4.mean_f(x, params), x[:,0]**params[0])
    assert_allclose(mf4.mean_deriv(x, np.zeros(1)), deriv_exp_0)
    assert_allclose(mf4.mean_deriv(x, oneparams), deriv_exp_1)
    assert_allclose(mf4.mean_deriv(x, params), deriv_exp_2)
    assert_allclose(mf4.mean_hessian(x, np.zeros(1)), hess_exp_0)
    assert_allclose(mf4.mean_hessian(x, oneparams), hess_exp_1)
    assert_allclose(mf4.mean_hessian(x, params), hess_exp_2)
    assert_allclose(mf4.mean_inputderiv(x, oneparams), inputderiv_exp_1)
    assert_allclose(mf4.mean_inputderiv(x, params), inputderiv_exp_2)

    deriv_fd = np.zeros((n_params, x.shape[0]))
    for i in range(n_params):
        dx_array = np.zeros(n_params)
        dx_array[i] = dx
        deriv_fd[i] = (mf4.mean_f(x, oneparams) - mf4.mean_f(x, oneparams - dx_array))/dx

    assert_allclose(mf4.mean_deriv(x, oneparams), deriv_fd, atol=1.e-5, rtol=1.e-5)

    deriv_fd = np.zeros((n_params, x.shape[0]))
    for i in range(n_params):
        dx_array = np.zeros(n_params)
        dx_array[i] = dx
        deriv_fd[i] = (mf4.mean_f(x, params) - mf4.mean_f(x, params - dx_array))/dx

    assert_allclose(mf4.mean_deriv(x, params), deriv_fd, atol=1.e-5, rtol=1.e-5)

    hess_fd = np.zeros((n_params, n_params, x.shape[0]))
    for i in range(n_params):
        for j in range(n_params):
            dx_array = np.zeros(n_params)
            dx_array[i] = dx
            hess_fd[i, j] = (mf4.mean_deriv(x, oneparams)[j] - mf4.mean_deriv(x, oneparams - dx_array)[j])/dx

    assert_allclose(mf4.mean_hessian(x, oneparams), hess_fd, atol=1.e-5, rtol=1.e-5)

    hess_fd = np.zeros((n_params, n_params, x.shape[0]))
    for i in range(n_params):
        for j in range(n_params):
            dx_array = np.zeros(n_params)
            dx_array[i] = dx
            hess_fd[i, j] = (mf4.mean_deriv(x, params)[j] - mf4.mean_deriv(x, params - dx_array)[j])/dx

    assert_allclose(mf4.mean_hessian(x, params), hess_fd, atol=1.e-5, rtol=1.e-5)

    inputderiv_fd = np.zeros((D, x.shape[0]))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (mf4.mean_f(x, oneparams) - mf4.mean_f(x - dx_array, oneparams))/dx

    assert_allclose(mf4.mean_inputderiv(x, oneparams), inputderiv_fd, atol=1.e-6, rtol=1.e-6)

    inputderiv_fd = np.zeros((D, x.shape[0]))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (mf4.mean_f(x, params) - mf4.mean_f(x - dx_array, params))/dx

    assert_allclose(mf4.mean_inputderiv(x, params), inputderiv_fd, atol=1.e-6, rtol=1.e-6)

    mf5 = Coefficient()**2

    assert mf5.get_n_params(x) == n_params

    assert_allclose(mf5.mean_f(x, params), params[0]**2)
    assert_allclose(mf5.mean_deriv(x, params), np.broadcast_to(2.*params[0], (n_params, x.shape[0])))
    assert_allclose(mf5.mean_hessian(x, params), np.broadcast_to(2., (n_params, n_params, x.shape[0])))
    assert_allclose(mf5.mean_inputderiv(x, params), np.zeros((x.shape[1], x.shape[0])))

    deriv_fd = np.zeros((n_params, x.shape[0]))
    for i in range(n_params):
        dx_array = np.zeros(n_params)
        dx_array[i] = dx
        deriv_fd[i] = (mf5.mean_f(x, params) - mf5.mean_f(x, params - dx_array))/dx

    assert_allclose(mf5.mean_deriv(x, params), deriv_fd, atol=1.e-5, rtol=1.e-5)

def test_MeanPower_specialcases():
    "test situations where MeanPower is badly behaved (complex, etc.) or could behave badly if not well implemented"

    # mean function should raise an error if not real

    mf = LinearMean(0)**2.1

    with pytest.raises(FloatingPointError):
        mf.mean_f(np.array([-2.]), np.zeros(0))

    # verify that if exponent has no parameters and x is negative still get correct functioning

    mf = LinearMean(0)**2.

    assert_allclose(mf.mean_deriv(np.array([-2.]), np.zeros(0)), -4.)

    # check that error raised for badly behaved derivative

    mf = LinearMean(0)**Coefficient()

    with pytest.raises(FloatingPointError):
        mf.mean_deriv(np.array([-2.]), np.array([2.]))

def test_MeanComposite(x, params, dx):
    "test the composite mean function"

    mf1 = LinearMean(0)*LinearMean(1)
    mf2 = Coefficient() + Coefficient()*LinearMean(0) + Coefficient()*LinearMean(0)*LinearMean(0)

    mf = mf2(mf1)

    n_params = 3

    assert mf.get_n_params(x) == n_params

    deriv_exp = np.zeros((n_params, x.shape[0]))
    deriv_exp[0,:] = 1.
    deriv_exp[1,:] = x[:,0]*x[:,1]
    deriv_exp[2,:] = x[:,0]**2*x[:,1]**2

    inputderiv_exp = np.zeros((x.shape[1], x.shape[0]))
    inputderiv_exp[0] = params[1]*x[:,1] + 2.*params[2]*x[:,0]*x[:,1]**2
    inputderiv_exp[1] = params[1]*x[:,0] + 2.*params[2]*x[:,0]**2*x[:,1]

    assert_allclose(mf.mean_f(x, params), params[0] + params[1]*x[:,0]*x[:,1]
                                                    + params[2]*x[:,0]**2*x[:,1]**2)
    assert_allclose(mf.mean_deriv(x, params), deriv_exp)
    assert_allclose(mf.mean_inputderiv(x, params), inputderiv_exp)

    with pytest.raises(NotImplementedError):
        mf.mean_hessian(x, params)

    mf3 = Coefficient() + Coefficient()*LinearMean(1) + Coefficient()*LinearMean(0)
    mf4 = mf3(mf1)

    with pytest.raises(IndexError):
        mf4.mean_f(x, params)

    with pytest.raises(IndexError):
        mf4.mean_deriv(x, params)

    with pytest.raises(IndexError):
        mf4.mean_inputderiv(x, params)

def test_PolynomialMean(x, dx):
    "test the polynomial mean function"

    degree = 2

    poly_mean = PolynomialMean(degree)
    n_params = 1 + x.shape[1]*degree

    assert poly_mean.degree == degree
    assert poly_mean.get_n_params(x) == n_params

    params = np.arange(1., 1.+float(n_params))

    mean_expected = np.array([(params[0] + params[1]*x[0, 0] + params[2]*x[0, 1] + params[3]*x[0, 2]
                                         + params[4]*x[0, 0]**2 + params[5]*x[0, 1]**2 + params[6]*x[0, 2]**2),
                              (params[0] + params[1]*x[1, 0] + params[2]*x[1, 1] + params[3]*x[1, 2]
                                         + params[4]*x[1, 0]**2 + params[5]*x[1, 1]**2 + params[6]*x[1, 2]**2)])
    deriv_expected = np.array([[1., 1.], x[:, 0], x[:, 1], x[:, 2],
                                         x[:, 0]**2, x[:, 1]**2, x[:, 2]**2])
    hess_expected = np.zeros((n_params, n_params, 2))
    inputderiv_expected = np.array([[params[1] + 2.*params[4]*x[0,0], params[1] + 2.*params[4]*x[1,0]],
                                    [params[2] + 2.*params[5]*x[0,1], params[2] + 2.*params[5]*x[1,1]],
                                    [params[3] + 2.*params[6]*x[0,2], params[3] + 2.*params[6]*x[1,2]]])

    assert_allclose(poly_mean.mean_f(x, params), mean_expected)
    assert_allclose(poly_mean.mean_deriv(x, params), deriv_expected)
    assert_allclose(poly_mean.mean_hessian(x, params), hess_expected)
    assert_allclose(poly_mean.mean_inputderiv(x, params), inputderiv_expected)

    deriv_fd = np.zeros((n_params, 2))
    for i in range(n_params):
        dx_array = np.zeros(n_params)
        dx_array[i] = dx
        deriv_fd[i] = (poly_mean.mean_f(x, params) - poly_mean.mean_f(x, params - dx_array))/dx

    assert_allclose(poly_mean.mean_deriv(x, params), deriv_fd)

    hess_fd = np.zeros((n_params, n_params, 2))
    for i in range(n_params):
        for j in range(n_params):
            dx_array = np.zeros(n_params)
            dx_array[j] = dx
            hess_fd[i, j] = (poly_mean.mean_deriv(x, params)[i] - poly_mean.mean_deriv(x, params - dx_array)[i])/dx

    assert_allclose(poly_mean.mean_hessian(x, params), hess_fd)

    D = 3

    inputderiv_fd = np.zeros((D, 2))
    for i in range(D):
        dx_array = np.zeros(D)
        dx_array[i] = dx
        inputderiv_fd[i] = (poly_mean.mean_f(x, params) - poly_mean.mean_f(x - dx_array, params))/dx

    assert_allclose(poly_mean.mean_inputderiv(x, params), inputderiv_fd, atol=1.e-5)

def test_PolynomialMean_failures(x):
    "test situations where PolynomialMean should raise an error"

    with pytest.raises(AssertionError):
        PolynomialMean(-1)

    degree = 2

    poly_mean = PolynomialMean(degree)

    with pytest.raises(AssertionError):
        poly_mean._check_inputs(x, np.ones(4))

    with pytest.raises(AssertionError):
        poly_mean._check_inputs(np.ones((2, 3, 2)), np.ones(7))


import numpy as np
import pytest
from numpy.testing import assert_allclose
from ..GaussianProcess import GaussianProcess
from ..MultiOutputGP import MultiOutputGP
from ..fitting import fit_GP_MAP, _fit_single_GP_MAP, _fit_MOGP_MAP

def test_fit_GP_MAP():
    "test the fit_GP_MAP function"

    # test correct basic functioning

    x = np.linspace(0., 1.)
    y = x**2

    gp = GaussianProcess(x, y)

    theta_exp = np.array([ 1.6030532031342832, -2.090511425471982 , -0.7803960307137198])
    logpost_exp = -296.0297245831661
    np.random.seed(4335)

    gp = fit_GP_MAP(gp)

    assert isinstance(gp, GaussianProcess)
    assert_allclose(gp.theta, theta_exp)
    assert_allclose(gp.current_logpost, logpost_exp)

    # same test, but pass args and kwargs rather than gp

    np.random.seed(4335)

    gp = fit_GP_MAP(x, y, mean="0.", use_patsy=False, method="L-BFGS-B")
    assert isinstance(gp, GaussianProcess)
    assert_allclose(gp.theta, theta_exp)
    assert_allclose(gp.current_logpost, logpost_exp)

    # minimization fails

    with pytest.raises(RuntimeError):
        fit_GP_MAP(gp, n_tries=1, theta0=-10000.*np.ones(3))

    gp = GaussianProcess(x, y, nugget=0.)

    with pytest.raises(RuntimeError):
        fit_GP_MAP(gp, n_tries=1)

    with pytest.raises(RuntimeError):
        fit_GP_MAP(gp, theta0 = np.array([800., 0., 0.]), n_tries=1)

    # bad inputs

    with pytest.raises(TypeError):
        fit_GP_MAP(x)

    with pytest.raises(TypeError):
        fit_GP_MAP()

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, n_tries=-1)

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, theta0=np.ones(1))


def test_fit_GP_MAP_MOGP():
    "test the fit_GP_MAP function with multiple outputs"

    x = np.linspace(0., 1.)
    y = np.zeros((2, 50))
    y[0] = x**2
    y[1] = 2. + x**3

    gp = MultiOutputGP(x, y)

    theta_exp = np.zeros((2, 3))
    theta_exp[0] = np.array([ 1.6030532031342832, -2.090511425471982 , -0.7803960307137198])
    theta_exp[1] = np.array([ 1.414112951818647 , -0.5974688573393576,  0.6857536842773265])
    logpost_exp = np.zeros(2)
    logpost_exp[0] = -296.0297245831661
    logpost_exp[1] = -250.06025683867367
    np.random.seed(4335)

    gp = fit_GP_MAP(gp, processes=1)

    assert isinstance(gp, MultiOutputGP)
    for i in range(2):
        assert_allclose(gp.emulators[i].theta, theta_exp[i])
        assert_allclose(gp.emulators[i].current_logpost, logpost_exp[i])

    # same test, but pass args and kwargs rather than gp

    np.random.seed(4335)

    gp = fit_GP_MAP(x, y, mean="0.", use_patsy=False, method="L-BFGS-B", processes=1)
    assert isinstance(gp, MultiOutputGP)
    for i in range(2):
        assert_allclose(gp.emulators[i].theta, theta_exp[i])
        assert_allclose(gp.emulators[i].current_logpost, logpost_exp[i])

    # pass processes argument

    np.random.seed(4335)

    gp = fit_GP_MAP(x, y, mean="0.", use_patsy=False, method="L-BFGS-B", processes=1)
    assert isinstance(gp, MultiOutputGP)

    # pass various theta0 arguments

    gp = fit_GP_MAP(x, y, theta0=np.zeros(3))

    gp = fit_GP_MAP(x, y, theta0=np.zeros((2, 3)))

    gp = fit_GP_MAP(x, y, theta0=[None, np.zeros(3)])

    # minimization fails

    with pytest.raises(RuntimeError):
        fit_GP_MAP(gp, n_tries=1, theta0=-10000.*np.ones(3))

    gp = MultiOutputGP(x, y, nugget=0.)

    with pytest.raises(RuntimeError):
        fit_GP_MAP(gp, n_tries=1)

    with pytest.raises(RuntimeError):
        fit_GP_MAP(gp, theta0 = np.array([800., 0., 0.]), n_tries=1)

    # bad inputs

    with pytest.raises(TypeError):
        fit_GP_MAP(x)

    with pytest.raises(TypeError):
        fit_GP_MAP()

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, n_tries=-1)

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, theta0=np.ones(1))

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, theta0=np.zeros((3, 3)))

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, theta0=np.zeros((2, 1)))

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, theta0=np.zeros((1, 1, 1)))

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, theta0=[None, None, None])

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, theta0=[np.zeros(1), np.zeros(3)], processes=1)

def test_fit_single_GP_MAP():
    "test the method to run the minimization algorithm on a GP class"

    x = np.linspace(0., 1.)
    y = x**2

    gp = GaussianProcess(x, y)

    theta_exp = np.array([ 1.6030532031342832, -2.090511425471982 , -0.7803960307137198])
    logpost_exp = -296.0297245831661
    np.random.seed(4335)

    gp = _fit_single_GP_MAP(gp)

    assert isinstance(gp, GaussianProcess)
    assert_allclose(gp.theta, theta_exp)
    assert_allclose(gp.current_logpost, logpost_exp)

    # minimization fails

    with pytest.raises(RuntimeError):
        _fit_single_GP_MAP(gp, n_tries=1, theta0=-10000.*np.ones(3))

    gp = GaussianProcess(x, y, nugget=0.)

    with pytest.raises(RuntimeError):
        _fit_single_GP_MAP(gp, n_tries=1)

    with pytest.raises(RuntimeError):
        _fit_single_GP_MAP(gp, theta0 = np.array([800., 0., 0.]), n_tries=1)

    # bad inputs

    with pytest.raises(AssertionError):
        _fit_single_GP_MAP(x)

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, n_tries=-1)

    with pytest.raises(AssertionError):
        fit_GP_MAP(gp, theta0=np.ones(1))

def test_fit_MOGP_MAP_MOGP():
    "test the fit_GP_MAP function with multiple outputs"

    x = np.linspace(0., 1.)
    y = np.zeros((2, 50))
    y[0] = x**2
    y[1] = 2. + x**3

    gp = MultiOutputGP(x, y)

    theta_exp = np.zeros((2, 3))
    theta_exp[0] = np.array([ 1.6030532031342832, -2.090511425471982 , -0.7803960307137198])
    theta_exp[1] = np.array([ 1.414112951818647 , -0.5974688573393576,  0.6857536842773265])
    logpost_exp = np.zeros(2)
    logpost_exp[0] = -296.0297245831661
    logpost_exp[1] = -250.06025683867367
    np.random.seed(4335)

    gp = _fit_MOGP_MAP(gp, processes=1)

    assert isinstance(gp, MultiOutputGP)
    for i in range(2):
        assert_allclose(gp.emulators[i].theta, theta_exp[i])
        assert_allclose(gp.emulators[i].current_logpost, logpost_exp[i])

    # pass processes argument

    np.random.seed(4335)

    gp = _fit_MOGP_MAP(gp, processes=1)
    assert isinstance(gp, MultiOutputGP)

    # pass various theta0 arguments

    gp = fit_GP_MAP(gp, theta0=np.zeros(3))

    gp = fit_GP_MAP(gp, theta0=np.zeros((2, 3)))

    gp = fit_GP_MAP(gp, theta0=[None, np.zeros(3)])

    # minimization fails

    with pytest.raises(RuntimeError):
        _fit_MOGP_MAP(gp, n_tries=1, theta0=-10000.*np.ones(3))

    gp = MultiOutputGP(x, y, nugget=0.)

    with pytest.raises(RuntimeError):
        _fit_MOGP_MAP(gp, n_tries=1)

    with pytest.raises(RuntimeError):
        _fit_MOGP_MAP(gp, theta0 = np.array([800., 0., 0.]), n_tries=1)

    # bad inputs

    with pytest.raises(AssertionError):
        _fit_MOGP_MAP(x)

    with pytest.raises(AssertionError):
        _fit_MOGP_MAP(gp, n_tries=-1)

    with pytest.raises(AssertionError):
        _fit_MOGP_MAP(gp, theta0=np.ones(1))

    with pytest.raises(AssertionError):
        _fit_MOGP_MAP(gp, theta0=np.zeros((3, 3)))

    with pytest.raises(AssertionError):
        _fit_MOGP_MAP(gp, theta0=np.zeros((2, 1)))

    with pytest.raises(AssertionError):
        _fit_MOGP_MAP(gp, theta0=np.zeros((1, 1, 1)))

    with pytest.raises(AssertionError):
        _fit_MOGP_MAP(gp, theta0=[None, None, None])

    with pytest.raises(AssertionError):
        _fit_MOGP_MAP(gp, theta0=[np.zeros(1), None], processes=1)

    with pytest.raises(AssertionError):
        _fit_MOGP_MAP(gp, processes=-1)


import numpy as np
from numpy.testing import assert_allclose
import pytest
from inspect import signature
import types
from ..ExperimentalDesign import LatinHypercubeDesign
from ..SequentialDesign import SequentialDesign, MICEDesign, MICEFastGP
from ..GaussianProcess import GaussianProcess
from ..fitting import fit_GP_MAP
from tempfile import TemporaryFile

def test_SequentialDesign_init():
    "test the init method of ExperimentalDesign"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed)
    assert type(sd.base_design).__name__ == 'LatinHypercubeDesign'
    assert sd.f == None
    assert sd.n_samples == None
    assert sd.n_init == 10
    assert sd.n_cand == 50
    assert sd.current_iteration == 0
    assert not sd.initialized
    assert sd.inputs == None
    assert sd.targets == None
    assert sd.candidates == None

    sd = SequentialDesign(ed, f, n_samples = 100, n_init = 20, n_cand = 100)
    assert type(sd.base_design).__name__ == 'LatinHypercubeDesign'
    assert callable(sd.f)
    assert len(signature(sd.f).parameters) == 1
    assert sd.n_samples == 100
    assert sd.n_init == 20
    assert sd.n_cand == 100
    assert sd.current_iteration == 0
    assert not sd.initialized
    assert sd.inputs == None
    assert sd.targets == None
    assert sd.candidates == None

def test_SequentialDesign_init_failures():
    "test cases where init should fail"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    with pytest.raises(TypeError):
        sd = SequentialDesign(3., f)

    with pytest.raises(TypeError):
        sd = SequentialDesign(ed, 3.)

    def f2(x, y):
        return np.array([1.])

    with pytest.raises(ValueError):
        sd = SequentialDesign(ed, f2)

    with pytest.raises(ValueError):
        sd = SequentialDesign(ed, f, n_samples = -1)

    with pytest.raises(ValueError):
        sd = SequentialDesign(ed, f, n_init = -1)

    with pytest.raises(ValueError):
        sd = SequentialDesign(ed, f, n_cand = -1)

def test_SequentialDesign_save_design():
    "test the save_design method"

    ed = LatinHypercubeDesign(3)
    sd = SequentialDesign(ed, n_init = 2, n_cand = 3)

    sd.inputs = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
    sd.targets = np.array([2., 4.])
    sd.candidates = np.array([[0.7, 0.8, 0.9], [0.15, 0.25, 0.35], [0.45, 0.55, 0.65]])

    with TemporaryFile() as tmp:
        sd.save_design(tmp)
        tmp.seek(0)
        design_file = np.load(tmp)
        assert_allclose(design_file['inputs'], np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]))
        assert_allclose(design_file['targets'], np.array([2., 4.]))
        assert_allclose(design_file['candidates'], np.array([[0.7, 0.8, 0.9],
                                                             [0.15, 0.25, 0.35], [0.45, 0.55, 0.65]]))

def test_SequentialDesign_load_design():
    "test the load_design method"

    ed = LatinHypercubeDesign(3)
    sd = SequentialDesign(ed, n_init = 2, n_cand = 3)

    with TemporaryFile() as tmp:
        np.savez(tmp, inputs=np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]),
                                      targets = np.array([2., 4.]),
                                      candidates = np.array([[0.7, 0.8, 0.9],
                                                             [0.15, 0.25, 0.35],
                                                             [0.45, 0.55, 0.65]]))
        tmp.seek(0)
        sd.load_design(tmp)

    assert_allclose(sd.inputs, np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]))
    assert_allclose(sd.targets, np.array([2., 4.]))
    assert_allclose(sd.candidates, np.array([[0.7, 0.8, 0.9],
                                             [0.15, 0.25, 0.35],
                                             [0.45, 0.55, 0.65]]))
    assert sd.initialized
    assert sd.current_iteration == 2

    # disagreement between base design and inputs

    ed = LatinHypercubeDesign(2)
    sd = SequentialDesign(ed, n_init = 2, n_cand = 3)

    with TemporaryFile() as tmp:
        np.savez(tmp, inputs=np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]),
                                      targets = np.array([2., 4.]),
                                      candidates = np.array([[0.7, 0.8, 0.9],
                                                             [0.15, 0.25, 0.35],
                                                             [0.45, 0.55, 0.65]]))
        tmp.seek(0)
        with pytest.raises(AssertionError):
            sd.load_design(tmp)

    # disagreement between shape of inputs and candidates

    ed = LatinHypercubeDesign(3)
    sd = SequentialDesign(ed, n_init = 2, n_cand = 3)

    with TemporaryFile() as tmp:
        np.savez(tmp, inputs=np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]),
                                      targets = np.array([2., 4.]),
                                      candidates = np.array([[0.7, 0.8],
                                                             [0.15, 0.25],
                                                             [0.45, 0.55]]))
        tmp.seek(0)
        with pytest.raises(AssertionError):
            sd.load_design(tmp)

    # error due to targets having a bad shape

    ed = LatinHypercubeDesign(3)
    sd = SequentialDesign(ed, n_init = 2, n_cand = 3)

    with TemporaryFile() as tmp:
        np.savez(tmp, inputs=np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]),
                                      targets = np.array([2., 4., 5.]),
                                      candidates = np.array([[0.7, 0.8],
                                                             [0.15, 0.25],
                                                             [0.45, 0.55]]))
        tmp.seek(0)
        with pytest.raises(AssertionError):
            sd.load_design(tmp)

    # error due to existing targets but no inputs

    ed = LatinHypercubeDesign(3)
    sd = SequentialDesign(ed, n_init = 2, n_cand = 3)

    with TemporaryFile() as tmp:
        np.savez(tmp, inputs = None, targets = np.array([2., 4., 5.]), candidates = None)
        tmp.seek(0)
        with pytest.raises(AssertionError):
            sd.load_design(tmp)

def test_SequentialDesign_has_function():
    "test has_function method"

    ed = LatinHypercubeDesign(3)

    sd = SequentialDesign(ed)
    assert not sd.has_function()

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f)
    assert sd.has_function()

def test_SequentialDesign_get_n_parameters():
    "test the get_n_parameters method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f)
    assert sd.get_n_parameters() == 3

def test_SequentialDesign_get_n_init():
    "test the get_n_init method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f, n_init = 20)
    assert sd.get_n_init() == 20

def test_SequentialDesign_get_n_samples():
    "test the get_n_cand method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f, n_samples = 20)
    assert sd.get_n_samples() == 20

def test_SequentialDesign_get_n_cand():
    "test the get_n_cand method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f, n_cand = 20)
    assert sd.get_n_cand() == 20


def test_SequentialDesign_get_current_iteration():
    "test the get_current_iteration method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f)
    assert sd.get_current_iteration() == 0

    sd.current_iteration = 20
    assert sd.get_current_iteration() == 20

def test_SequentialDesign_get_inputs():
    "test the get_inputs method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f)
    assert sd.get_inputs() == None

    sd.inputs = np.zeros((3, 4))
    assert_allclose(sd.get_inputs(), np.zeros((3, 4)))

def test_SequentialDesign_get_targets():
    "test the get_targets method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f)
    assert sd.get_targets() == None

    sd.targets = np.zeros(4)
    assert_allclose(sd.get_targets(), np.zeros(4))

def test_SequentialDesign_get_candidates():
    "test the get_candidates method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f)
    assert sd.get_candidates() == None

    sd.candidates = np.zeros((3, 4))
    assert_allclose(sd.get_candidates(), np.zeros((3, 4)))

def test_SequentialDesign_get_base_design():
    "test the get_base_design method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f)
    assert sd.get_base_design() == "LatinHypercubeDesign"

def test_SequentialDesign_generate_initial_design():
    "test the generate_initial_design method"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.array([1.])

    sd = SequentialDesign(ed, f, n_init = 4)
    initial_design_expected = np.array([[0.9660431763890672, 0.2080126306969736, 0.2576380063570568],
                                        [0.0684779421445063, 0.9308367720360009, 0.1428493015158686],
                                        [0.6345029195085983, 0.6651343562344474, 0.8827198350687029],
                                        [0.4531112960399023, 0.3977273628763245, 0.5867585643640021]])
    initial_design_actual = sd.generate_initial_design()
    assert_allclose(initial_design_actual, initial_design_expected)
    assert_allclose(sd.inputs, initial_design_expected)
    assert sd.current_iteration == 4

    sd = SequentialDesign(ed, f, n_init = 4)
    sd.run_initial_design()

    with pytest.raises(AssertionError):
        sd.generate_initial_design()

def test_SequentialDesign_set_initial_targets():
    "test the set_initial_targets method"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    sd = SequentialDesign(ed, f, n_init = 4)
    initial_design_expected = np.array([[0.9660431763890672, 0.2080126306969736, 0.2576380063570568],
                                        [0.0684779421445063, 0.9308367720360009, 0.1428493015158686],
                                        [0.6345029195085983, 0.6651343562344474, 0.8827198350687029],
                                        [0.4531112960399023, 0.3977273628763245, 0.5867585643640021]])
    targets_expected = np.array([np.sum(i) for i in initial_design_expected])
    sd.generate_initial_design()
    sd.set_initial_targets(targets_expected)
    assert_allclose(sd.targets, np.reshape(targets_expected, (4,)))
    assert sd.initialized

    sd = SequentialDesign(ed, f, n_init = 4)
    targets_expected = np.array([np.sum(i) for i in initial_design_expected])
    sd.generate_initial_design()
    sd.set_initial_targets(targets_expected)
    assert_allclose(sd.targets, targets_expected)
    assert sd.initialized

    sd = SequentialDesign(ed, f, n_init = 4)
    with pytest.raises(ValueError):
        sd.set_initial_targets(targets_expected)

    sd = SequentialDesign(ed, f, n_init = 4)
    sd.generate_initial_design()
    with pytest.raises(AssertionError):
        sd.set_initial_targets(np.zeros((2, 4)))

    sd = SequentialDesign(ed, f, n_init = 4)
    sd.inputs = np.zeros((5,2))
    with pytest.raises(AssertionError):
        sd.set_initial_targets(targets_expected)

def test_SequentialDesign_run_initial_design():
    "test method to run initial design"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    sd = SequentialDesign(ed, f, n_init = 4)
    initial_design_expected = np.array([[0.9660431763890672, 0.2080126306969736, 0.2576380063570568],
                                        [0.0684779421445063, 0.9308367720360009, 0.1428493015158686],
                                        [0.6345029195085983, 0.6651343562344474, 0.8827198350687029],
                                        [0.4531112960399023, 0.3977273628763245, 0.5867585643640021]])
    targets_expected = np.array([np.sum(i) for i in initial_design_expected])
    sd.run_initial_design()
    assert_allclose(sd.inputs, initial_design_expected)
    assert_allclose(sd.targets, targets_expected)
    assert sd.initialized
    assert sd.current_iteration == 4

    sd = SequentialDesign(ed)
    with pytest.raises(AssertionError):
        sd.run_initial_design()

def test_SequentialDesign_generate_candidates():
    "test the _generate_candidates method"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    sd = SequentialDesign(ed, f, n_cand = 4)
    candidates_expected = np.array([[0.9660431763890672, 0.2080126306969736, 0.2576380063570568],
                                    [0.0684779421445063, 0.9308367720360009, 0.1428493015158686],
                                    [0.6345029195085983, 0.6651343562344474, 0.8827198350687029],
                                    [0.4531112960399023, 0.3977273628763245, 0.5867585643640021]])
    sd._generate_candidates()
    assert_allclose(sd.candidates, candidates_expected)

def test_SequentialDesign_eval_metric():
    "test the _eval_metric method"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    sd = SequentialDesign(ed, f)

    with pytest.raises(NotImplementedError):
        sd._eval_metric()

def test_SequentialDesign_estimate_next_target():
    "test the estimate next point method of sequential design"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    sd = SequentialDesign(ed, f)

    with pytest.raises(NotImplementedError):
        sd._estimate_next_target(np.array([0., 1., 2.]))

def test_SequentialDesign_get_batch_points():
    "test the get_batch_points method of a MICE design"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    def tmp_eval_metric(self):
        return 0

    def tmp_estimate_next_target(self, next_point):
        return np.array([1.])

    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd._estimate_next_target = types.MethodType(tmp_estimate_next_target, sd)

    sd.run_initial_design()

    next_points = sd.get_batch_points(2)

    assert_allclose(next_points, np.array([[3.9602716910300234e-01, 4.3469440375712098e-02, 9.3294684823072194e-01],
                                           [0.1314127131166828, 0.3850568631590907, 0.2234836206262954]]))
    assert_allclose(sd.inputs, np.array([[0.9660431763890672, 0.2080126306969736, 0.2576380063570568],
                                         [0.0684779421445063, 0.9308367720360009, 0.1428493015158686],
                                         [0.6345029195085983, 0.6651343562344474, 0.8827198350687029],
                                         [0.4531112960399023, 0.3977273628763245, 0.5867585643640021],
                                         [3.9602716910300234e-01, 4.3469440375712098e-02, 9.3294684823072194e-01],
                                         [0.1314127131166828, 0.3850568631590907, 0.2234836206262954]]))
    assert_allclose(sd.targets, np.array([1.4316938134430974, 1.1421640156963757, 2.1823571108117488,
                                          1.437597223280229 ]))
    assert sd.current_iteration == 4

    # check raises error if number of batch points is negative

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    sd.run_initial_design()

    with pytest.raises(AssertionError):
        sd.get_batch_points(-1)

def test_SequentialDesign_get_next_point():
    "test the get_next_point method"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    def tmp_eval_metric(self):
        return 0

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_initial_design()
    next_point = sd.get_next_point()

    next_point_expected = np.array([3.9602716910300234e-01, 4.3469440375712098e-02, 9.3294684823072194e-01])
    inputs_expected = np.array([[0.9660431763890672, 0.2080126306969736, 0.2576380063570568],
                                [0.0684779421445063, 0.9308367720360009, 0.1428493015158686],
                                [0.6345029195085983, 0.6651343562344474, 0.8827198350687029],
                                [0.4531112960399023, 0.3977273628763245, 0.5867585643640021],
                                [3.9602716910300234e-01, 4.3469440375712098e-02, 9.3294684823072194e-01]])
    candidates_expected = np.array([[3.9602716910300234e-01, 4.3469440375712098e-02, 9.3294684823072194e-01],
                                    [6.7353171430888004e-01, 3.2967090024809237e-01, 4.7561098969781079e-01],
                                    [9.9391552790043791e-01, 5.0015522522360367e-01, 2.4850731127782361e-01],
                                    [6.6098271480205528e-04, 9.7616245971529891e-01, 7.4815518971467765e-01]])
    assert_allclose(next_point, next_point_expected)
    assert_allclose(sd.inputs, inputs_expected)
    assert_allclose(sd.candidates, candidates_expected)


    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    sd.generate_initial_design()
    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)

    with pytest.raises(ValueError):
        next_point = sd.get_next_point()

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    with pytest.raises(ValueError):
        next_point = sd.get_next_point()

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)
    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_initial_design()
    sd.inputs = np.zeros((5,3))

    with pytest.raises(AssertionError):
        next_point = sd.get_next_point()

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)
    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_initial_design()
    sd.targets = np.zeros((5,3))

    with pytest.raises(AssertionError):
        next_point = sd.get_next_point()

def test_SequentialDesign_set_batch_targets():
    "test the set_batch_targets method"

    ed = LatinHypercubeDesign(3)

    sd = SequentialDesign(ed, n_init = 4, n_cand = 4)

    sd.inputs = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9],
                          [0.15, 0.25, 0.35], [0.45, 0.55, 0.65], [0.75, 0.85, 0.95]])
    sd.targets = np.array([1., 2., 3., 4.])
    sd.current_iteration = 4
    sd.initialized = True

    sd.set_batch_targets(np.array([5., 6.]))

    assert_allclose(sd.targets, np.array([1., 2., 3., 4., 5., 6.]))
    assert sd.current_iteration == 6

    # failure if new_targets wrong shape

    ed = LatinHypercubeDesign(3)

    sd = SequentialDesign(ed, n_init = 4, n_cand = 4)

    sd.inputs = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9],
                          [0.15, 0.25, 0.35], [0.45, 0.55, 0.65], [0.75, 0.85, 0.95]])
    sd.targets = np.array([1., 2., 3., 4.])

    with pytest.raises(AssertionError):
        sd.set_batch_targets(np.array([5., 6., 7.]))

def test_SequentialDesign_set_next_target():
    "test the set_next_target method"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    def tmp_eval_metric(self):
        return 0

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_initial_design()
    next_point = sd.get_next_point()
    new_target = np.reshape(np.array(np.sum(next_point)), (1,))

    inputs_expected = np.array([[0.9660431763890672, 0.2080126306969736, 0.2576380063570568],
                                [0.0684779421445063, 0.9308367720360009, 0.1428493015158686],
                                [0.6345029195085983, 0.6651343562344474, 0.8827198350687029],
                                [0.4531112960399023, 0.3977273628763245, 0.5867585643640021],
                                [3.9602716910300234e-01, 4.3469440375712098e-02, 9.3294684823072194e-01]])
    targets_expected = np.array([np.sum(i) for i in inputs_expected])

    sd.set_next_target(new_target)
    assert_allclose(sd.targets, targets_expected)
    assert sd.current_iteration == 5


    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_initial_design()
    next_point = sd.get_next_point()

    with pytest.raises(AssertionError):
        sd.set_next_target(np.zeros((2,)))


    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    sd.generate_initial_design()
    sd.inputs = np.zeros((5,3))
    next_target = np.zeros(1)

    with pytest.raises(ValueError):
        sd.set_next_target(next_target)

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    with pytest.raises(ValueError):
        sd.set_next_target(next_target)

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)
    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_initial_design()

    with pytest.raises(AssertionError):
        sd.set_next_target(next_target)

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)
    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_initial_design()
    sd.get_next_point()
    sd.targets = np.zeros((5,3))

    with pytest.raises(AssertionError):
        sd.set_next_target(next_target)

def test_SequentialDesign_run_next_point():
    "test the run_next_point method"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    def tmp_eval_metric(self):
        return 0

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)
    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_initial_design()
    sd.run_next_point()

    inputs_expected = np.array([[0.9660431763890672, 0.2080126306969736, 0.2576380063570568],
                                [0.0684779421445063, 0.9308367720360009, 0.1428493015158686],
                                [0.6345029195085983, 0.6651343562344474, 0.8827198350687029],
                                [0.4531112960399023, 0.3977273628763245, 0.5867585643640021],
                                [3.9602716910300234e-01, 4.3469440375712098e-02, 9.3294684823072194e-01]])
    targets_expected = np.array([np.sum(i) for i in inputs_expected])

    assert_allclose(sd.inputs, inputs_expected)
    assert_allclose(sd.targets, targets_expected)
    assert sd.current_iteration == 5

    sd = SequentialDesign(ed)
    with pytest.raises(AssertionError):
        sd.run_next_point()


def test_SequentialDesign_run_sequential_design():
    "test the run_sequential_design method"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    def tmp_eval_metric(self):
        return 0

    inputs_expected = np.array([[0.9660431763890672, 0.2080126306969736, 0.2576380063570568],
                                [0.0684779421445063, 0.9308367720360009, 0.1428493015158686],
                                [0.6345029195085983, 0.6651343562344474, 0.8827198350687029],
                                [0.4531112960399023, 0.3977273628763245, 0.5867585643640021],
                                [3.9602716910300234e-01, 4.3469440375712098e-02, 9.3294684823072194e-01],
                                [0.1314127131166828, 0.3850568631590907, 0.2234836206262954],
                                [0.1648353557812244, 0.0994384529732742, 0.4715221513612055],
                                [0.8739475357732106, 0.058541390000348 , 0.3103313392459021]])
    targets_expected = np.array([np.sum(i) for i in inputs_expected])

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)
    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_sequential_design(4)

    assert_allclose(sd.inputs, inputs_expected)
    assert_allclose(sd.targets, targets_expected)
    assert sd.current_iteration == 8

    np.random.seed(74632)

    sd = SequentialDesign(ed, f, n_samples = 4, n_init = 4, n_cand = 4)
    sd._eval_metric = types.MethodType(tmp_eval_metric, sd)
    sd.run_sequential_design()

    assert_allclose(sd.inputs, inputs_expected)
    assert_allclose(sd.targets, targets_expected)
    assert sd.current_iteration == 8

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)

    with pytest.raises(ValueError):
        sd.run_sequential_design()

    sd = SequentialDesign(ed)
    with pytest.raises(AssertionError):
        sd.run_sequential_design()

    sd = SequentialDesign(ed, f, n_init = 4, n_cand = 4)
    with pytest.raises(AssertionError):
        sd.run_sequential_design(n_samples = -1)

def test_SequentialDesign_str():
    "test string method of sequential design"

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    expected_string = ""
    expected_string += "SequentialDesign with\n"
    expected_string += "LatinHypercubeDesign base design\n"
    expected_string += "None total samples\n"
    expected_string += "10 initial points\n"
    expected_string += "50 candidate points\n"
    expected_string += "0 current samples\n"
    expected_string += "current inputs: None\n"
    expected_string += "current targets: None"

    sd = SequentialDesign(ed)
    assert str(sd) == expected_string

    expected_string = ""
    expected_string += "SequentialDesign with\n"
    expected_string += "LatinHypercubeDesign base design\n"
    expected_string += "a bound simulator function\n"
    expected_string += "10 total samples\n"
    expected_string += "5 initial points\n"
    expected_string += "10 candidate points\n"
    expected_string += "0 current samples\n"
    expected_string += "current inputs: None\n"
    expected_string += "current targets: None"

    sd = SequentialDesign(ed, f, n_samples = 10, n_init = 5, n_cand = 10)
    assert str(sd) == expected_string

def test_MICEDesign_init():
    "test the init method of MICEDesign"

    ed = LatinHypercubeDesign(3)

    md = MICEDesign(ed)

    assert type(md.base_design).__name__ == 'LatinHypercubeDesign'
    assert md.f == None
    assert md.n_samples == None
    assert md.n_init == 10
    assert md.n_cand == 50
    assert md.current_iteration == 0
    assert not md.initialized
    assert md.inputs == None
    assert md.targets == None
    assert md.candidates == None
    assert md.nugget == "adaptive"
    assert_allclose(md.nugget_s, 1.)

    def f(x):
        return np.array([1.])

    md = MICEDesign(ed, f, 20, 5, 40, 1.e-12, 0.1)

    assert type(md.base_design).__name__ == 'LatinHypercubeDesign'
    assert callable(md.f)
    assert len(signature(md.f).parameters) == 1
    assert md.n_samples == 20
    assert md.n_init == 5
    assert md.n_cand == 40
    assert md.current_iteration == 0
    assert not md.initialized
    assert md.inputs == None
    assert md.targets == None
    assert md.candidates == None
    assert_allclose(md.nugget, 1.e-12)
    assert_allclose(md.nugget_s, 0.1)

def test_MICEDesign_init_failures():
    "test occasions where MICE Design should fail upon initializing"

    ed = LatinHypercubeDesign(3)

    with pytest.raises(ValueError):
        md = MICEDesign(ed, nugget = -1.)

    with pytest.raises(ValueError):
        md = MICEDesign(ed, nugget_s = -1.)

def test_MICEDesign_get_nugget():
    "test the get_nugget method of MICE Design"

    ed = LatinHypercubeDesign(3)

    md = MICEDesign(ed)

    assert md.get_nugget() == "adaptive"

    md = MICEDesign(ed, nugget = 1.)

    assert_allclose(md.get_nugget(), 1.)

def test_MICEDesign_get_nugget_s():
    "test the get_nugget_s method of MICE Design"

    ed = LatinHypercubeDesign(3)

    md = MICEDesign(ed)

    assert_allclose(md.get_nugget_s(), 1.)

def test_MICEDesign_estimate_next_target():
    "test the estimate next target method for a MICE design"

    ed = LatinHypercubeDesign(3)
    md = MICEDesign(ed)

    md.gp = GaussianProcess(np.array([[1., 2., 3.], [4., 5., 6]]), np.array([2., 4.]))
    md.gp.theta = np.zeros(md.gp.n_params)

    assert_allclose(np.array([0.0018237589305011]), md._estimate_next_target(np.zeros(3)))

    with pytest.raises(AssertionError):
        md._estimate_next_target(np.zeros(2))

def test_MICEDesign_MICE_criterion():
    "test the function to compute the MICE criterion"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    md = MICEDesign(ed, f, n_init = 4, n_cand = 4)

    md.run_initial_design()
    md._generate_candidates()

    md.gp = GaussianProcess(md.get_inputs(), md.get_targets())
    md.gp = fit_GP_MAP(md.gp)

    md.gp_fast = MICEFastGP(md.candidates, np.ones(4), nugget=1.)
    md.gp_fast.theta = md.gp.theta

    metric = md._MICE_criterion(0)

    metric_expected = 0.0899338596342571

    assert_allclose(metric, metric_expected, rtol = 1.e-6)

    with pytest.raises(AssertionError):
        metric = md._MICE_criterion(-1)

    with pytest.raises(AssertionError):
        metric = md._MICE_criterion(5)

def test_MICEDesign_eval_metric():
    "test the _eval_metric method of MICE Design"

    np.random.seed(74632)

    ed = LatinHypercubeDesign(3)

    def f(x):
        return np.sum(x)

    md = MICEDesign(ed, f, n_init = 4, n_cand = 4)

    md.run_initial_design()
    md._generate_candidates()

    best_point = md._eval_metric()

    best_point_expected = 0

    assert best_point == best_point_expected

def test_MICEFastGP():
    "test the correction formula for the modified GP for Fast MICE"

    gp = MICEFastGP(np.reshape([1., 2., 3., 4], (4, 1)), [1., 1., 1., 1.])
    gp.theta = np.array([0., -1., 0.])
    result = gp.fast_predict(3)
    result_expected = 0.191061906777163

    assert_allclose(result, result_expected)

    with pytest.raises(AssertionError):
        result = gp.fast_predict(-1)

    with pytest.raises(AssertionError):
        result = gp.fast_predict(5)


import numpy as np
import scipy.stats
from inspect import signature

class ExperimentalDesign(object):
    """
    Base class representing a generic one-shot design of experiments with uncorrelated parameters
    
    This class provides the base implementation for a class for designing experiments to sample
    the parameter space of a complex model. The parameter space can be specified in a variety
    of ways, but essentially the user must provide a Probability Point Function (PPF, or inverse 
    of the Cumulative Distribution Function) for each input parameter. Each PPF function takes a 
    single numeric input and maps from the interval :math:`[0,1]` to the desired parameter
    distribution value for a given parameter, and each parameter has a separate function describing 
    its distribution. Note that this makes the assumption of no correlations between any of the
    parameter values (a future version may implement an experimental design where there are such
    parameter correlations). Once the design is initialized, a desired number of samples can be
    drawn from the design, returning an array holding the desired number of samples from the
    parameter space.
    
    Internally, the class holds the set of PPFs for all of the parameter values, and samples are
    drawn by calling the ``sample`` method. To draw the samples, a specific method ``_draw_samples``
    must be defined that generates a series of points in the :math:`[0,1]^n` hypercube, where
    :math:`n` is the number of paramters. This set of samples from the hypercube is then mapped to
    the parameter space using the given PPF functions. Thus, defining a new design protocol only
    requires defining a new ``_draw_samples`` method and redefining the ``__init__`` method to set
    the internal ``method`` attribute. By default, no ``_draw_samples`` method is defined, so the
    base ``ExperimentalDesign`` class is only intended to be used to define new protocols (trying 
    to sample from an ``ExperimentalDesign`` instance will return a ``NotImplementedError``).
    """
    def __init__(self, *args):
        """
        Create a new instance of an experimental design
        
        Creates a new instance of a design of experiments, which draws samples from the parameter
        space of a complex model. It is often used to generate data for a Gaussian Process emulator
        to fit the outputs of the complex model. This is a base class that does not implement the
        method for sampling from the distribution; to use an experimental design in practice you
        should use one of the derived classes provided or create your own.
        
        The experimental design can be initialized in several ways depending on the arguments
        provided, ranging from the simplest to the most complicated.
        
        1. Provide an integer ``n`` indicating the number of input parameters. If this is used to
           create an instance, it is assumed that all parameters are unformly distributed over
           the :math:`n`-dimensional hypercube.
        2. Provide an integer ``n`` and a tuple ``(a, b)`` of length 2 containing two numeric values
           (where :math:`a < b`). In this case, all parameters are assumed to be uniformly distributed
           over the interval :math:`[a,b]`.
        3. Provide an integer ``n`` and a function that takes a single numeric input in the interval
           :math:`[0,1]` and maps it to the parameter space. In this case, all parameters are assumed
           to follow the provided Probability Point Function.
        4. Provide a list of tuples of length 2 containing numeric values (as above, the first number
           must smaller than the second number). The design then assumes that the number of parameters
           is the length of the list, and each parameter follows a uniform distribution with the bounds
           given by the respective tuple in the given list.
        5. Provide a list of functions taking a single input (as above, each function must map the
           interval :math:`[0,1]` to the parameter space). The number of parameters in the design is the
           length of the list, and the given PPF functions define the parameter space for each input.
        
        More concretely, if one input parameter is given, you may initilize the class in any of the
        following ways:
        
        :param n_parameters: Integer specifying the number of parameters (must be positive). The
                             design will sample each parameter over the interval :math:`[0,1]`
        :type n_parameters: int
        
        or
        
        :param bounds_list: List of tuples containing two numeric values, each of which has the
                           smaller number first. Each parameter then takes a uniform distribution
                           with bounds given by each tuple.
        :type bounds_list: list
        
        or
        
        :param ppf_list: List of functions or other callable, each of which accepts one argument
                         and maps the interval :math:`[0,1]` to the parameter space. Each parameter
                         follows the distribution given by the respective PPF function.
        :type ppf_list: list
        
        and if two input parameters are given:
        
        :param n_parameters: Integer specifying the number of parameters (must be positive). The
                             design will sample each parameter over the interval :math:`[0,1]`
        :type n_parameters: int
        :param bounds: Tuple or other iterable containing two numeric values, where the smaller
                       number must come first. Each parameter then takes a uniform distribution
                       with bounds given by the numbers provided in the tuple.
        :type bounds: tuple
        
        or
        
        :param n_parameters: Integer specifying the number of parameters (must be positive). The
                             design will sample each parameter over the interval :math:`[0,1]`
        :type n_parameters: int
        :param ppf: Function or other callable, which accepts one argument and maps the interval
                    :math:`[0,1]` to the parameter space. Each parameter follows the distribution
                    given by the PPF function.
        :type ppf: function
        
        The ``scipy.stats`` package provides implementations of a wide range of distributions, with
        pre-defined PPF functions. See the Scipy user manual for more details. Note that in order to
        get a compatible PPF function that only takes a single input, you will need to set any parameters
        needed to define the distribution.
        
        Internally, the class defines any PPF functions based on the input data and collects all of
        the PPF functions in a list. The class also contains information on the method used to draw
        samples from the design.
        
        To create a usable implementation based on this class, the user must define the method
        ``_draw_samples``, which takes a positive integer input ``n_samples`` and draws ``n_samples``
        from the :math:`[0,1]^n` hypercube, where :math:`n` is the number of parameters. The user must
        also modify the ``method`` attribute of the design in order to have the ``__str__`` method work
        correctly. All other functionality should not require any changes from the base class.
        """
        
        if len(args) == 1:
            try:
                n_parameters = int(args[0])
                bounds = None
            except TypeError:
                try:
                    n_parameters = len(list(args[0]))
                    bounds = list(args[0])
                except TypeError:
                    raise TypeError("bad input type for ExperimentalDesign")
        elif len(args) == 2:
            try:
                n_parameters = int(args[0])
            except TypeError:
                raise TypeError("bad input type for ExperimentalDesign")
                
            if callable(args[1]):
                bounds = args[1]
            else:
                try:
                    bounds = list(args[1])
                    try:
                        if (len(bounds) == 2 and isinstance(float(bounds[0]), float) and isinstance(float(bounds[1]), float)):
                            if float(bounds[1]) <= float(bounds[0]):
                                raise ValueError("bad value for parameter bounds in ExperimentalDesign")
                            bounds = (float(bounds[0]), float(bounds[1]))
                    except TypeError:
                        pass
                except TypeError:
                    raise TypeError("bad input type for ExperimentalDesign")
        else:
            raise ValueError("bad inputs for ExperimentalDesign")
                
        if n_parameters <= 0:
            raise ValueError("number of parameters must be positive in Experimental Design")

        self.n_parameters = n_parameters
        
        if bounds is None:
            self.distributions = [scipy.stats.uniform(loc = 0., scale = 1.).ppf]*n_parameters
        elif isinstance(bounds, tuple):
            self.distributions = [scipy.stats.uniform(loc = bounds[0], scale = bounds[1]-bounds[0]).ppf]*n_parameters
        elif callable(bounds):
            if len(signature(bounds).parameters) == 1:
                self.distributions = [bounds]*n_parameters
            else:
                raise ValueError("PPF distribution provided must accept a single argument")
        else: # bounds is a list
            if not len(bounds) == n_parameters:
                raise ValueError("list of parameter distributions must have the same length")
            self.distributions = []
            for item in bounds:
                if callable(item):
                    if len(signature(item).parameters) == 1:
                        self.distributions.append(item)
                    else:
                        raise ValueError("PPF distribution provided must accept a single argument")
                else:
                    try:
                        if (len(item) == 2 and isinstance(float(item[0]), float) and isinstance(float(item[1]), float)):
                            if float(item[1]) <= float(item[0]):
                                raise ValueError("bad value for parameter bounds in ExperimentalDesign")
                            self.distributions.append(scipy.stats.uniform(loc = float(item[0]), 
                                                                          scale = float(item[1])- float(item[0])).ppf)
                        else:
                            raise ValueError("bounds for each parameter must be a tuple of two floats")
                    except TypeError:
                        raise TypeError("bounds for each parameter must be a tuple of two floats")
            
        
    def get_n_parameters(self):
        """
        Returns number of parameters in the experimental design
        
        This method returns the number of parameters in the experimental design. This is set when
        initializing the object, an cannot be modified.
        
        :returns: Number of parameters in the experimental design.
        :rtype: int
        """
        return self.n_parameters
        
    def get_method(self):
        """
        Returns the method used to draw samples from the design
        
        This method returns the method used to draw samples from the experimental design. The base
        class does not implement a method, so if you try to call this on the base class the code
        will raise a ``NotImplementedError``. When deriving new designs from the base class,
        the method should be set when calling the ``__init__`` method.
        
        :returns: Method used to draw samples from the design.
        :rtype: str
        """
        try:
            return self.method
        except AttributeError:
            raise NotImplementedError("base class of ExperimentalDesign does not implement a method")
    
    def _draw_samples(self, n_samples):
        """
        Low level method for drawing random samples from a design
        
        This method implements the low level method for sampling from a design. All samples drawn
        using this method will lie in the :math:`[0,1]^n` hypercube, where :math:`n` is the number
        of parameters in the design. Once these are drawn, the actual parameter values are obtained
        by transforming the samples using the provided PPF functions.
        
        This method is not implemented for the base class. Derived classes will only need to
        implement this method to get a working design, as all other pieces of code should be mostly
        identical.
        
        :param n_samples: Number of samples to be drawn from the design (must be a positive integer)
        :type n_samples: int
        :returns: Samples drawn from the :math:`[0,1]^n` hypercube as a numpy array with shape
                  ``(n_samples, n_parameters)``
        :rtype: ndarray
        """
        raise NotImplementedError
    
    def sample(self, n_samples):
        """
        Draw parameter samples from the experimental design
        
        This method implements drawing parameter samples from the experimental design. The method does
        this by calling the ``_draw_samples`` method to obtain samples from the :math:`[0,1]^n` hypercube,
        where :math:`n` is the number of parameters. The ``sample``method then transforms these samples
        drawn from the low level method to the actual parameter values using the PPF functions provided
        when initilizing the object. Note that this method also checks that all parameter values are
        finite; if any ``NaN`` values are returned, an error will be raised.
        
        Note that by implementing the sampling in this way, modifications to the method to draw samples
        using a different protocol only needs to change the ``_draw_samples`` method. This makes it
        simpler to define new designs, as only a single method needs to be altered.
        
        :param n_samples: Number of samples to be drawn from the design (must be a positive integer)
        :type n_samples: int
        :returns: Samples drawn from the design parameter space as a numpy array with shape
                  ``(n_samples, n_parameters)``
        :rtype: ndarray
        """
        
        n_samples = int(n_samples)
        assert n_samples > 0, "number of samples must be positive"
        
        sample_values = np.zeros((n_samples, self.get_n_parameters()))
        random_draws = self._draw_samples(n_samples)
        
        assert np.all(random_draws >= 0.) and np.all(random_draws <= 1.), "error in generating random samples"
        
        for (dist, index) in zip(self.distributions, range(self.get_n_parameters())):
            try:
                sample_values[:,index] = dist(random_draws[:,index])
            except:
                for sample_index in range(n_samples):
                    sample_values[sample_index, index] = dist(random_draws[sample_index,index])
        
        assert np.all(np.isfinite(sample_values)), "error due to non-finite values of parameters"
        
        return sample_values
        
    def __str__(self):
        """
        Returns a string representation of the ExperimentalDesign object
        
        This method returns a string representation of the Experimental Design object. If a specific
        method is set for a derived class from the base class, the string will include this method
        in the string output. Otherwise, a generic string will be returned.
        
        :returns: String representation of the object
        :rtype: str
        """
        try:
            method = self.get_method()+" "
        except NotImplementedError:
            method = ""
        return method+"Experimental Design with "+str(self.get_n_parameters())+" parameters"
        
class MonteCarloDesign(ExperimentalDesign):
    """
    Class representing a one-shot design of experiments with uncorrelated parameters using Monte
    Carlo Sampling
    
    This class provides an implementation for a class for designing experiments to sample
    the parameter space of a complex model using random Monte Carlo sampling. The parameter space can
    be specified in a variety of ways, but essentially the user must provide a Probability Point
    Function (PPF, or inverse of the Cumulative Distribution Function) for each input parameter. Each
    PPF function takes a single numeric input and maps from the interval :math:`[0,1]` to the desired
    parameter distribution value for a given parameter, and each parameter has a separate function
    describing its distribution. Note that this makes the assumption of no correlations between any of
    the parameter values (a future version may implement an experimental design where there are such
    parameter correlations). Once the design is initialized, a desired number of samples can be
    drawn from the design, returning an array holding the desired number of samples from the
    parameter space.
    
    Internally, the class holds the set of PPFs for all of the parameter values, and samples are
    drawn by calling the ``sample`` method. To draw the samples, the ``_draw_samples`` is used
    to generate a series of points in the :math:`[0,1]^n` hypercube using random Monte Carlo sampling,
    where :math:`n` is the number of paramters. This set of samples from the hypercube is then mapped to
    the parameter space using the given PPF functions.
    
    This design does not attempt to uniformly sample the space, but rather just makes random draws from
    the parameter distributions. For this reason, small designs using this method may not sample the
    parameter space very efficiently. However, generating a large number of samples using this protocol
    can be done very efficiently, as drawing a sample only requires generating a series of pseduorandom
    numbers. 
    """
    def __init__(self, *args):
        """
        Create a new instance of a Monte Carlo experimental design
    
        Creates a new instance of a Monte Carlo design of experiments, which draws samples randomly 
        from the parameter space of a complex model. It can be used to generate data for a Gaussian
        Process emulator to fit the outputs of the complex model. Because the samples are drawn
        randomly, small designs may not sample the space very well, but the samples can be drawn
        quickly as they are entirely random.
    
        The experimental design can be initialized in several ways depending on the arguments
        provided, ranging from the simplest to the most complicated.
    
        1. Provide an integer ``n`` indicating the number of input parameters. If this is used to
           create an instance, it is assumed that all parameters are unformly distributed over
           the :math:`n`-dimensional hypercube.
        2. Provide an integer ``n`` and a tuple ``(a, b)`` of length 2 containing two numeric values
           (where :math:`a < b`). In this case, all parameters are assumed to be uniformly distributed
           over the interval :math:`[a,b]`.
        3. Provide an integer ``n`` and a function that takes a single numeric input in the interval
           :math:`[0,1]` and maps it to the parameter space. In this case, all parameters are assumed
           to follow the provided Probability Point Function.
        4. Provide a list of tuples of length 2 containing numeric values (as above, the first number
           must smaller than the second number). The design then assumes that the number of parameters
           is the length of the list, and each parameter follows a uniform distribution with the bounds
           given by the respective tuple in the given list.
        5. Provide a list of functions taking a single input (as above, each function must map the
           interval :math:`[0,1]` to the parameter space). The number of parameters in the design is the
           length of the list, and the given PPF functions define the parameter space for each input.
    
        More concretely, if one input parameter is given, you may initilize the class in any of the
        following ways:
    
        :param n_parameters: Integer specifying the number of parameters (must be positive). The
                             design will sample each parameter over the interval :math:`[0,1]`
        :type n_parameters: int
    
        or
    
        :param bounds_list: List of tuples containing two numeric values, each of which has the
                           smaller number first. Each parameter then takes a uniform distribution
                           with bounds given by each tuple.
        :type bounds_list: list
    
        or
    
        :param ppf_list: List of functions or other callable, each of which accepts one argument
                         and maps the interval :math:`[0,1]` to the parameter space. Each parameter
                         follows the distribution given by the respective PPF function.
        :type ppf_list: list
    
        and if two input parameters are given:
    
        :param n_parameters: Integer specifying the number of parameters (must be positive). The
                             design will sample each parameter over the interval :math:`[0,1]`
        :type n_parameters: int
        :param bounds: Tuple or other iterable containing two numeric values, where the smaller
                       number must come first. Each parameter then takes a uniform distribution
                       with bounds given by the numbers provided in the tuple.
        :type bounds: tuple
    
        or
    
        :param n_parameters: Integer specifying the number of parameters (must be positive). The
                             design will sample each parameter over the interval :math:`[0,1]`
        :type n_parameters: int
        :param ppf: Function or other callable, which accepts one argument and maps the interval
                    :math:`[0,1]` to the parameter space. Each parameter follows the distribution
                    given by the PPF function.
        :type ppf: function
    
        The ``scipy.stats`` package provides implementations of a wide range of distributions, with
        pre-defined PPF functions. See the Scipy user manual for more details. Note that in order to
        get a compatible PPF function that only takes a single input, you will need to set any parameters
        needed to define the distribution.
    
        Internally, the class defines any PPF functions based on the input data and collects all of
        the PPF functions in a list. The class also contains information on the method used to draw
        samples from the design.
        """
        
        self.method = "Monte Carlo"
        super().__init__(*args)
        
    def _draw_samples(self, n_samples):
        """
        Low level method for drawing random samples from a Monte Carlo design
        
        This method implements the low level method for sampling from a Monte Carlo design. All samples
        drawn using this method are random points in the :math:`[0,1]^n` hypercube, where :math:`n` is
        the number of parameters in the design. Once these are drawn, the actual parameter values are
        obtained by transforming the samples using the provided PPF functions in the ``sample`` method.
        
        :param n_samples: Number of samples to be drawn from the design (must be a positive integer)
        :type n_samples: int
        :returns: Random Monte Carlo samples drawn from the :math:`[0,1]^n` hypercube as a numpy
                  array with shape ``(n_samples, n_parameters)``
        :rtype: ndarray
        """
        
        n_samples = int(n_samples)
        assert n_samples > 0, "number of samples must be positive"
        
        return np.random.random((n_samples, self.get_n_parameters()))
        
        
class LatinHypercubeDesign(ExperimentalDesign):
    """
    Class representing a one-shot design of experiments with uncorrelated parameters using Latin
    Hypercube Sampling
    
    This class provides an implementation for a class for designing experiments to sample
    the parameter space of a complex model using Latin Hypercube sampling. The parameter space can
    be specified in a variety of ways, but essentially the user must provide a Probability Point
    Function (PPF, or inverse of the Cumulative Distribution Function) for each input parameter. Each
    PPF function takes a single numeric input and maps from the interval :math:`[0,1]` to the desired
    parameter distribution value for a given parameter, and each parameter has a separate function
    describing its distribution. Note that this makes the assumption of no correlations between any of
    the parameter values (a future version may implement an experimental design where there are such
    parameter correlations). Once the design is initialized, a desired number of samples can be
    drawn from the design, returning an array holding the desired number of samples from the
    parameter space.
    
    Internally, the class holds the set of PPFs for all of the parameter values, and samples are
    drawn by calling the ``sample`` method. To draw the samples, the ``_draw_samples`` is used
    to generate a series of points in the :math:`[0,1]^n` hypercube using Latin Hypercube sampling,
    where :math:`n` is the number of paramters. This set of samples from the Latin Hypercube is then
    mapped to the parameter space using the given PPF functions.
    
    Unlike Monte Carlo sampling, Latin Hypercube designs attempt to sample more uniformly from the
    parameter space. Latin Hypercube sampling ensures that each sample is drawn from a different
    part of the space for each parameter. For example, if four samples are drawn, then for each
    parameter, one sample is guaranteed to be drawn from each quartile of the distribution. This
    ensures a more uniform sampling when compared on Monte Carlo sampling, but requires slightly
    more computation to generate the samples. Note however, that for very large numbers of parameters,
    Latin Hypercubes still may not sample very efficiently. This is due to the fact that the size of
    the parameter space grows exponentially with the number of dimensions, so a fixed number of
    samples will sample the space more poorly as the number of parameters increases.
    """
    def __init__(self, *args):
        """
        Create a new instance of a Latin Hypercube experimental design
    
        Creates a new instance of a Latin Hypercube design of experiments, which draws samples 
        from the parameter space of a complex model in a more uniform fashion when compared to
        random Monte Carlo sampling. It can be used to generate data for a Gaussian
        Process emulator to fit the outputs of the complex model. Because the samples are drawn
        more uniformly than in Monte Carlo sampling, these designs may perform better in high
        dimensional parameter spaces.
    
        The experimental design can be initialized in several ways depending on the arguments
        provided, ranging from the simplest to the most complicated.
    
        1. Provide an integer ``n`` indicating the number of input parameters. If this is used to
           create an instance, it is assumed that all parameters are unformly distributed over
           the :math:`n`-dimensional hypercube.
        2. Provide an integer ``n`` and a tuple ``(a, b)`` of length 2 containing two numeric values
           (where :math:`a < b`). In this case, all parameters are assumed to be uniformly distributed
           over the interval :math:`[a,b]`.
        3. Provide an integer ``n`` and a function that takes a single numeric input in the interval
           :math:`[0,1]` and maps it to the parameter space. In this case, all parameters are assumed
           to follow the provided Probability Point Function.
        4. Provide a list of tuples of length 2 containing numeric values (as above, the first number
           must smaller than the second number). The design then assumes that the number of parameters
           is the length of the list, and each parameter follows a uniform distribution with the bounds
           given by the respective tuple in the given list.
        5. Provide a list of functions taking a single input (as above, each function must map the
           interval :math:`[0,1]` to the parameter space). The number of parameters in the design is the
           length of the list, and the given PPF functions define the parameter space for each input.
    
        More concretely, if one input parameter is given, you may initilize the class in any of the
        following ways:
    
        :param n_parameters: Integer specifying the number of parameters (must be positive). The
                             design will sample each parameter over the interval :math:`[0,1]`
        :type n_parameters: int
    
        or
    
        :param bounds_list: List of tuples containing two numeric values, each of which has the
                           smaller number first. Each parameter then takes a uniform distribution
                           with bounds given by each tuple.
        :type bounds_list: list
    
        or
    
        :param ppf_list: List of functions or other callable, each of which accepts one argument
                         and maps the interval :math:`[0,1]` to the parameter space. Each parameter
                         follows the distribution given by the respective PPF function.
        :type ppf_list: list
    
        and if two input parameters are given:
    
        :param n_parameters: Integer specifying the number of parameters (must be positive). The
                             design will sample each parameter over the interval :math:`[0,1]`
        :type n_parameters: int
        :param bounds: Tuple or other iterable containing two numeric values, where the smaller
                       number must come first. Each parameter then takes a uniform distribution
                       with bounds given by the numbers provided in the tuple.
        :type bounds: tuple
    
        or
    
        :param n_parameters: Integer specifying the number of parameters (must be positive). The
                             design will sample each parameter over the interval :math:`[0,1]`
        :type n_parameters: int
        :param ppf: Function or other callable, which accepts one argument and maps the interval
                    :math:`[0,1]` to the parameter space. Each parameter follows the distribution
                    given by the PPF function.
        :type ppf: function
    
        The ``scipy.stats`` package provides implementations of a wide range of distributions, with
        pre-defined PPF functions. See the Scipy user manual for more details. Note that in order to
        get a compatible PPF function that only takes a single input, you will need to set any parameters
        needed to define the distribution.
    
        Internally, the class defines any PPF functions based on the input data and collects all of
        the PPF functions in a list. The class also contains information on the method used to draw
        samples from the design.
        """
        
        self.method = "Latin Hypercube"
        super().__init__(*args)
    
    def _draw_samples(self, n_samples):
        """
        Low level method for drawing random samples from a Latin Hypercube design
        
        This method implements the low level method for sampling from a Latin Hypercube design. All
        samples drawn using this method are random points in the :math:`[0,1]^n` hypercube, where
        :math:`n` is the number of parameters in the design. Once these are drawn, the actual parameter
        values are obtained by transforming the samples using the provided PPF functions in the ``sample``
        method. A Latin Hypercube design guarantees that each sample is drawn from a unique portion
        of the parameter space. For example, if 10 samples are drawn from a design, then for each
        parameter the ten samples will each come from a different decile of the distribution.
        
        :param n_samples: Number of samples to be drawn from the design (must be a positive integer)
        :type n_samples: int
        :returns: Random Monte Carlo samples drawn from the :math:`[0,1]^n` hypercube as a numpy
                  array with shape ``(n_samples, n_parameters)``
        :rtype: ndarray
        """
        
        n_samples = int(n_samples)
        assert n_samples > 0, "number of samples must be positive"
        
        n_parameters = self.get_n_parameters()

        random_samples = np.reshape(np.tile(np.arange(n_samples, dtype = np.float)/float(n_samples), n_parameters),
                                            (n_parameters, n_samples))
                    
        for row in random_samples:
            np.random.shuffle(row)
            
        random_samples =  np.transpose(random_samples) + np.random.random((n_samples, n_parameters))/float(n_samples)
        
        assert np.all(random_samples >= 0.) and np.all(random_samples <= 1.), "error in generating latin hypercube samples"
        
        return random_samples


import numpy as np
import pytest
from numpy.testing import assert_allclose
from ..Priors import NormalPrior, GammaPrior, InvGammaPrior
from scipy.stats import norm, gamma, invgamma

@pytest.fixture
def dx():
    return 1.e-6

def test_NormalPrior(dx):
    "test the NormalPrior class"

    normprior = NormalPrior(2., 3.)

    assert_allclose(normprior.logp(0.5), np.log(norm.pdf(0.5, loc=2., scale=3.)))

    assert_allclose(normprior.dlogpdtheta(0.5),
                    (normprior.logp(0.5) - normprior.logp(0.5 - dx))/dx, atol=1.e-7, rtol=1.e-7)

    assert_allclose(normprior.d2logpdtheta2(0.5),
                    (normprior.dlogpdtheta(0.5) - normprior.dlogpdtheta(0.5 - dx))/dx, atol=1.e-7, rtol=1.e-7)

def test_GammaPrior(dx):
    "test the GammaPrior class"

    gprior = GammaPrior(2., 3.)

    assert_allclose(gprior.logp(0.5), np.log(gamma.pdf(np.exp(0.5), 2., scale=3.)))

    assert_allclose(gprior.dlogpdtheta(0.5),
                    (gprior.logp(0.5) - gprior.logp(0.5 - dx))/dx, atol=1.e-6, rtol=1.e-6)

    assert_allclose(gprior.d2logpdtheta2(0.5),
                    (gprior.dlogpdtheta(0.5) - gprior.dlogpdtheta(0.5 - dx))/dx, atol=1.e-6, rtol=1.e-6)

def test_InvGammaPrior(dx):
    "test the InvGammaPrior class"

    igprior = InvGammaPrior(2., 3.)

    assert_allclose(igprior.logp(0.5), np.log(invgamma.pdf(np.exp(0.5), 2., scale=3.)))

    assert_allclose(igprior.dlogpdtheta(0.5),
                    (igprior.logp(0.5) - igprior.logp(0.5 - dx))/dx, atol=1.e-6, rtol=1.e-6)

    assert_allclose(igprior.d2logpdtheta2(0.5),
                    (igprior.dlogpdtheta(0.5) - igprior.dlogpdtheta(0.5 - dx))/dx, atol=1.e-6, rtol=1.e-6)



import numpy as np
from scipy.spatial.distance import cdist
from inspect import signature
from mogp_emulator.ExperimentalDesign import ExperimentalDesign
from mogp_emulator.GaussianProcess import GaussianProcess
from mogp_emulator.fitting import fit_GP_MAP
from numpy.linalg import LinAlgError

class SequentialDesign(object):
    """
    Base class representing a sequential experimental design

    This class provides the base implementation of a class for designing experiments sequentially. This
    means that rather than picking all simulation points in a single step, the points are selected one
    by one, taking into account the information obtained by determining the true parameter value at each
    design point when selecting the next one. Sequential designs can be very useful when running expensive,
    high-dimensional simulations to ensure that a limited computational budget is used effectvely.

    Instead of choosing all points at once, which is the case in a one-shot design, a sequential design
    does some additional computation work at each step to more carefully choose the next point. This means
    that sequential designs are better suited for very expensive simulations, where the additional
    cost of choosing the next point is small compared to the overall computational cost of running
    the simulations.

    A sequential design is built on top of a base design (which must be a subclass of the
    ``ExperimentalDesign`` class. In addition to the base design, the class must contain information on
    how many points are used in the initial design (i.e. the number of starting points used before starting
    the sequential steps in the design) and the number of candidate points that are considered during each
    iteration. Optionally, a function for evaluating the actual simulation can be optionally bound to the
    class instance, which allows the entire design process to be automated. If such a function is not
    provided, then the steps to run the design must be carried out manually, with the evaluated
    simulation values provided to the class at the end of each simulation in order to determine the
    next point.

    To use the base class to create an experimental design, a new subclass must be created that provides
    a method ``_eval_metric``, which considers all candidate points and returns the index of the best
    candidate. Otherwise, all other code provided here allows for a generic sequential design to be
    easily run and managed.
    """
    def __init__(self, base_design, f = None, n_samples = None, n_init = 10, n_cand = 50):
        """
        Create a new instance of a sequential experimental design

        Creates a new instance of a sequential experimental design, which sequentially chooses
        points to be evaluated from a complex simulation function. It is often used for
        expensive computational models, where the cost of running a single evaluation is
        large and must be done in series due to computational limitations, and thus some
        additional computation done at each step to select new points is small compared
        to the overall cost of running a single simulation.

        Sequential designs require specifying a base design using a subclass of ``ExperimentalDesign``
        as well as information on the number of points to use in each step in the design
        process. Additionally, the function to evaluated can be bound to the class to allow
        automatic evaluation of the function at each step.

        :param base_design: Base one-shot experimental design (must be a subclass of
                            ``ExperimentalDesign``). This contains the information on the
                            parameter space to be sampled.
        :type base_design: ExperimentalDesign
        :param f: Function to be evaluated for the design. Must take all parameter values as a single
                  input array and return a single float or an array of length 1
        :type f: function or other callable
        :param n_samples: Number of sequential design points to be drawn. If specified, this must be
                          a non-negative integer. Note that this is in addition to the number of initial
                          points, meaning that the total design size will be ``n_samples + n_init``.
                          This can also be specified when running the full design. This parameter is
                          optional, and defaults to ``None`` (meaning the number of samples is set when
                          running the design, or that samples will be added manually).
        :type n_samples: int or None
        :param n_init: Number of points in the inital design before the sequential steps begin. Must
                       be a positive integer. Optional, default value is 10.
        :type n_init: int
        :param n_cand: Number of candidates to consider at each sequential design step. Must be a positive
                       integer. Optional, default value is 50.
        """

        if not isinstance(base_design, ExperimentalDesign):
            raise TypeError("base design must be a one-shot experimental design")

        if not f is None:
            if not callable(f):
                raise TypeError("simulator f must be a function or other callable")

            if not len(signature(f).parameters) == 1:
                raise ValueError("simulator f must accept all parameters as a single input array")

        if (not n_samples is None) and int(n_samples) < 0:
            raise ValueError("number of samples must be nonzero")

        if int(n_init) <= 0:
            raise ValueError("number of initial design points must be positive")

        if int(n_cand) <= 0:
            raise ValueError("number of candidate design points must be positive")

        self.base_design = base_design
        self.f = f
        if n_samples is None:
            self.n_samples = None
        else:
            self.n_samples = int(n_samples)
        self.n_init = int(n_init)
        self.n_cand = int(n_cand)

        self.current_iteration = 0
        self.initialized = False
        self.inputs = None
        self.targets = None
        self.candidates = None

    def save_design(self, filename):
        """
        Save current state of the sequential design

        Saves the current state of the sequential design by writing the current
        values of ``inputs``, ``targets``, and ``candidates`` to file as a ``.npz``
        file. To re-load a saved design, use the ``load_design`` method.

        Note that this method only dumps the arrays holding the inputs, targets, and
        candidates to a ``.npz`` file. It does not ensure that the function or base
        design are consistent, so it is up to the user to ensure that the new design
        parameters are the same as the parameters for the old one.

        :param filename: Filename or file object where design will be saved
        :type filename: str or file
        :returns: None
        """

        design_dict = {}
        design_dict['inputs'] = self.inputs
        design_dict['targets'] = self.targets
        design_dict['candidates'] = self.candidates

        np.savez(filename, **design_dict)

    def load_design(self, filename):
        """
        Load previously saved sequential design

        Loads a previously saved sequential design from file. Loads the arrays for
        ``inputs``, ``targets``, and ``candidates`` from file and sets other internal
        data to be consistent. It performs a few checks for consistency to ensure
        that the loaded design is compatible with the selected parameters, however,
        it does not completely check everything for consistency (in particular, it does
        not make any attempt to ensure that the exact base design or function are
        identical to what was previously used). It is up to the user to ensure that
        these are consistent with the previous instance of the design.

        :param filename: Filename or file object from which the design will be loaded
        :type filename: str or file
        :returns: None
        """

        design_file = np.load(filename, allow_pickle=True)

        self.inputs = np.array(design_file['inputs'])
        if np.all(self.inputs) == None:
            self.inputs = None

        self.targets = np.array(design_file['targets'])
        if np.all(self.targets) == None:
            self.targets = None

        self.candidates = np.array(design_file['candidates'])
        if np.all(self.candidates) == None:
            self.candidates = None

        # perform some checks (note this is not exhaustive)

        if self.inputs is None:
            assert self.targets is None, "Cannot have targets without corresponding inputs"
        else:
            if not self.targets is None:
                assert self.targets.ndim == 1, "bad number of dimensions for targets"
                assert self.targets.shape[0] <= self.inputs.shape[0], "targets cannot be longer than inputs"
                self.initialized = True
                self.current_iteration = self.targets.shape[0]
            assert self.get_n_parameters() == self.inputs.shape[1], "Bad shape for inputs"
            if self.inputs.shape[1] < self.n_init:
                print("n_init greater than number of inputs, changing n_init")
                self.n_init = self.inputs.shape[1]

        if not self.candidates is None:
            assert self.get_n_parameters() == self.candidates.shape[1], "Bad shape for candidates"
            if self.candidates.shape[0] != self.n_cand:
                print("shape of candidates differs from n_cand, candidates will be overridden")

    def has_function(self):
        """
        Determines if class contains a function for running the simulator

        This method checks to see if a function has been provided for running the simulation.

        :returns: Whether or not the design has a bound function for evaluting the simulation.
        :rtype: bool
        """
        return (not self.f is None)

    def get_n_parameters(self):
        """
        Get number of parameters in design

        Returns the number of parameters in the design (note that this is specified in the base
        design that must be provided when initializing the class instance).

        :returns: Number of parameters in the design
        :rtype: int
        """

        return self.base_design.get_n_parameters()

    def get_n_init(self):
        """
        Get number of initial design points

        Returns the number of initial design points used before beginning the sequential design
        steps. Note that this means that the total number of samples to be drawn for the design
        is ``n_init + n_samples``.

        :returns: Number of initial design points
        :rtype: int
        """

        return self.n_init

    def get_n_samples(self):
        """
        Get number of sequential design points

        Returns the number of sequential design points used in the sequential design steps. This
        parameter can be ``None`` to indicate that the number of samples will be specified when
        running the design, or that the samples will be updated manually. Note that the total number
        of samples to be drawn for the design is ``n_init + n_samples``.

        :returns: Number of sequential design points
        :rtype: int
        """

        return self.n_samples

    def get_n_cand(self):
        """
        Get number of candidate design points

        Returns the number of candidate design points used in each sequential design step. Candidates
        are re-drawn at each step, so this number of points will be drawn each time and all points
        will be considered at each iteration.

        :returns: Number of candidate design points
        :rtype: int
        """

        return self.n_cand

    def get_current_iteration(self):
        """
        Get number of current iteration in the experimental design

        Returns the current iteration during the sequential design process. This is mostly useful
        if the sequential design is being updated manually to know the current iteration.

        :returns: Current iteration number
        :rtype: int
        """

        return self.current_iteration

    def get_inputs(self):
        """
        Get current design input points

        Returns a numpy array holding the current design points. The array is 2D and has shape
        ``(current_iteration, n_parameters)`` (i.e. it is resized after each iteration when a new
        design point is chosen).

        :returns: Current value of the design inputs
        :rtype: ndarray
        """

        return self.inputs

    def get_targets(self):
        """
        Get current design target points

        Returns a numpy array holding the current target points. The array is 1D and has shape
        ``(current_iteration,)`` (i.e. it is resized after each iteration when a new target point
        is added). Note that simulation outputs must be a single number, so if considering a
        simulation has multiple outputs, the user must decide how to combine them to form the
        relevant target value for deciding which point to simulate next.

        :returns: Current value of the target inputs
        :rtype: ndarray
        """

        return self.targets

    def get_candidates(self):
        """
        Get current candidate design input points

        Returns a numpy array holding the current candidate design points. The array is 2D and
        has shape ``(n_cand, n_parameters)``. It always has the same size once it is initialized,
        but the values will change acros iterations as new candidate points are considered at
        each iteration.

        :returns: Current value of the candidate design inputs
        :rtype: ndarray
        """

        return self.candidates

    def get_base_design(self):
        """
        Get type of base design

        Returns the type of the base design. The base design must be a subclass of ``ExperimentalDesign``,
        but any one-shot design method can be used to generate the initial design and the candidates.

        :returns: Base design type as a string
        :rtype: str
        """

        return type(self.base_design).__name__

    def generate_initial_design(self):
        """
        Create initial design

        Method to set the initial design inputs. Generates the desired number of points for the initial
        design by drawing from the base design. Method sets the ``inputs`` attribute of the
        ``SequentialDesign`` instance, but also returns the initial design as a numpy array if the
        simulations are to be run manually. This method can be run repeatedly to draw different
        initial designs if the initial target values have not been set, but once the targets have been
        set the method will not overwrite them to prevent corruption of the design.

        :returns: Initial design points, a 2D numpy array with shape ``(n_init, n_parameters)``
        :rtype: ndarray
        """

        assert not self.initialized, "initial design has already been created"

        self.inputs = self.base_design.sample(self.n_init)
        self.current_iteration = self.n_init
        return self.inputs

    def set_initial_targets(self, targets):
        """
        Set initial design target values

        Method to set the initial design targets. Generates the desired number of points for the initial
        design by drawing from the base design. Method sets the ``inputs`` attribute of the
        ``SequentialDesign`` instance, but also returns the initial design as a numpy array if the
        simulations are to be run manually. This method can be run repeatedly to draw different
        initial designs if the initial target values have not been set, but once the targets have been
        set the method will not overwrite them to prevent corruption of the design.

        Target values must be an array with length ``(n_init,)``, with values obtained by running
        the initial design through the simulation. Note that this means the initial design must
        be created prior to running this method -- if this method is called prior to
        ``generate_initial_design``, the code will raise an error.

        :param targets: Initial value of targets, must be a 1D numpy array with shape ``(n_init,)``
        :type targets: ndarray
        :returns: None
        :rtype: None
        """

        if self.inputs is None:
            raise ValueError("Initial design has not been generated")
        else:
            assert self.inputs.shape == (self.n_init, self.get_n_parameters()), "inputs have not been initialized correctly"

        targets = np.atleast_1d(np.squeeze(np.array(targets)))
        assert np.array(targets).shape == (self.n_init,), "initial targets must have shape (n_init,)"

        self.targets = np.array(targets)
        self.initialized = True

    def run_initial_design(self):
        """
        Run initial design

        Method to run the initial design by generating the initial design, evaluating the function on
        all design points, and setting the target values. Note that this requires having a bound function
        to the class in order to evaluate the design points internally. It is a shortcut to running
        ``generate_initial_design``, evaluating the initial design points, and then using
        ``set_initial_targets`` to set the target values, with some additional checks along the way.

        If the initial design has already been fully run, this method will raise an error as the
        method to generate the initial design checks this prior to overwriting the initial targets.
        Note also that this method checks that the outputs of the bound function match up with
        the expected array sizes and that all outputs are finite before updating the initial targets.

        :returns: None
        :rtype: None
        """


        assert self.has_function(), "Design must have a bound function to use run_initial_design"

        inputs = self.generate_initial_design()
        targets = np.full((self.n_init,), np.nan)

        for i in range(self.n_init):
            targets[i] = np.array(self.f(inputs[i,:]))

        assert np.all(np.isfinite(targets)), "error in initializing sequential design, function outputs may not be the correct shape"
        self.set_initial_targets(targets)

    def _generate_candidates(self):
        """
        Generate candidates for next iteration

        Internal method for generating candidates for the next iteration of the sequential design.
        Draws the desired number of points from the base design and sets the internal ``candidates``
        attribute to the resuting candidate design points.

        :returns: None
        :rtype: None
        """

        self.candidates = self.base_design.sample(self.n_cand)

    def _eval_metric(self):
        """
        Evaluate metric for selecting next point

        Apply the metric used for sequential design to all candidate points and returns the index of the
        best candidate. This is not implemented for the base implementation, and is the only method
        that should need to be updated in order to create a new type of sequential design.

        :returns: Index of best candidate from the possible next design points. Must be an integer
                  0 <= index < n_cand
        :rtype: int
        """
        raise NotImplementedError("Base class for Sequential Design does not implement an evaluation metric")

    def _estimate_next_target(self, next_point):
        """
        Estimate value of simulator for a point in a Sequential design

        This method is used for the batch version of a sequential design. Instead of updating
        the targets with the known solution, this method is used to estimate the function
        instead. This is method-specific, so this is not defined for the base class but instead
        should be defined in the subclass. Returns an array of length 1 holding the prediction.

        :param next_point: Input to be simulated. Must be an array of shape ``(n_parameters,)``
        :type next_point: ndarray
        :returns: Estimated simulation value for the given input as an array of length 1
        :rtype: ndarray
        """
        raise NotImplementedError("_estimate_next_point not implemented for base SequentialDesign")

    def get_batch_points(self, n_points):
        """
        Batch version of get_next_point for a Sequential Design

        This method returns a batch of design points to run from a Sequential Design. This is
        useful if simulations can be run in parallel, which speeds up the ability to
        generate designs efficiently. The method simply calls ``get_next_point`` the
        required number of times, but rather than using the true value of the simulation
        it instead substitutes the predicted value that is method-specific. This can be
        implemented in a subclass by defining the method ``_estimate_next_target``.

        :param n_points: Size of batch to generate for the next set of simulation points.
                         This parameter determines the shape of the output array. Must
                         be a positive integer.
        :type n_points: int
        :returns: Set of batch points chosen using the batch version of the design
                  as a numpy array with shape ``(n_points, n_parameters)``
        :rtype: ndarray
        """

        assert n_points > 0, "n_points must be positive"

        batch_points = np.zeros((n_points, self.get_n_parameters()))

        for i in range(n_points):
            batch_points[i] = self.get_next_point()
            next_target = self._estimate_next_target(batch_points[i])
            self.set_next_target(next_target)

        self.current_iteration = self.current_iteration - n_points
        new_targets = np.array(self.targets[:self.current_iteration])
        self.targets = np.array(new_targets)

        return batch_points

    def get_next_point(self):
        """
        Evaluate candidates to determine next point

        Public method for determining the next point in the design. Internally, it checks that the inputs
        and target arrays are as expected for correctly drawing a new point, generates prospective candidates,
        and then evaluates them using the desired metric in order to select the best one. It updates the
        ``inputs`` array and returns the next point to be evaluated as a 1D numpy array of length
        ``n_parameters``.

        :returns: Next design point, a 1D numpy array of length ``n_parameters``
        :rtype: ndarray
        """

        if self.inputs is None:
            raise ValueError("Initial design has not been generated")
        else:
            assert self.inputs.shape == (self.current_iteration, self.get_n_parameters()), "inputs have not been correctly updated"

        if self.targets is None:
            raise ValueError("Initial targets have not been generated")
        else:
            assert self.targets.shape == (self.current_iteration,), "targets have not been correctly updated"

        self._generate_candidates()
        next_index = self._eval_metric()

        new_inputs = np.empty((self.current_iteration + 1, self.get_n_parameters()))
        new_inputs[:-1, :] = self.inputs

        next_point = self.candidates[next_index,:]
        new_inputs[-1,:] = next_point

        self.inputs = np.array(new_inputs)

        return next_point

    def set_batch_targets(self, new_targets):
        """
        Batch version of set_next_target for a Sequential Design

        This method updates the targets array for a batch set of simulations. The input
        array must have shape ``(n_points,)``, where ``n_points`` is the number of points
        selected when calling ``get_batch_points``. Disagreement between these two values
        will result in an error.

        :param new_targets: Array holding results from the simulations. Must be an array
                            of shape ``(n_points,)``, where ``n_points`` is set when
                            calling ``get_batch_points``
        :type new_targets: ndarray
        :returns: None
        """
        if self.inputs is None:
            raise ValueError("Initial design has not been generated")
        else:
            n_points = self.inputs.shape[0] - self.current_iteration
            assert self.inputs.shape == (self.current_iteration + n_points, self.get_n_parameters()), "inputs have not been correctly updated"

        if self.targets is None:
            raise ValueError("Initial targets have not been generated")
        else:
            assert self.targets.shape == (self.current_iteration,), "targets have not been correctly updated"

        new_targets = np.atleast_1d(np.array(new_targets))
        new_targets = np.reshape(new_targets, (len(new_targets),))
        assert new_targets.shape == (n_points,), "new targets must have length n_points"

        updated_targets = np.empty((self.current_iteration + n_points,))
        updated_targets[:-n_points] = self.targets
        updated_targets[-n_points:] = np.array(new_targets)

        self.targets = np.array(updated_targets)
        self.current_iteration = self.current_iteration + n_points

    def set_next_target(self, target):
        """
        Set value of next target

        Updates the target array with the correct value (from running the actual simulation) of the
        latest design point determined using ``get_next_point``. The target input must be a float
        or an array of length 1. The code internally checks the inputs and targets for any problems
        that may have occurred in updating them correctly, and if all is well then updates the
        target array and increments the number of iterations. If the design has not been
        correctly initialized, or ``get_next_point`` has not been previously run, this method
        will raise an error.

        :param target: New target value found from evaluating the simulation on the latest design
                       point found from the ``get_next_point`` method.
        :type target: float or length 1 array
        :returns: None
        :rtype: None
        """

        if self.inputs is None:
            raise ValueError("Initial design has not been generated")
        else:
            assert self.inputs.shape == (self.current_iteration + 1, self.get_n_parameters()), "inputs have not been correctly updated"

        if self.targets is None:
            raise ValueError("Initial targets have not been generated")
        else:
            assert self.targets.shape == (self.current_iteration,), "targets have not been correctly updated"

        target = np.atleast_1d(np.array(target))
        target = np.reshape(target, (len(target),))
        assert target.shape == (1,), "new target must have length 1"

        new_targets = np.empty((self.current_iteration + 1,))
        new_targets[:-1] = self.targets
        new_targets[-1] = np.array(target)

        self.targets = np.array(new_targets)
        self.current_iteration = self.current_iteration + 1

    def run_next_point(self):
        """
        Perform one iteration of the sequential design process

        Method for performing an iteration of the sequential design process. This is a shortcut for
        generating and evaluating the candidates to find the best next design point, evaluating
        the function on the next point, and then updating the targets array with the value.
        This requires a function be bound to the class instance to automatically run the
        simulation. This will also automatically update the ``current_iteration`` attribute,
        which can be used to determine the number of sequential design steps that have been run.

        :returns: None
        :rtype: None
        """

        assert self.has_function(), "Design must have a bound function to use run_next_point"

        next_point = self.get_next_point()
        next_target = np.array(self.f(next_point))
        self.set_next_target(next_target)

    def run_sequential_design(self, n_samples = None):
        """
        Run the entire sequential design

        Method to run all steps of the sequential design process. Note that the class instance must
        have a bound function for evaluating the design points to run all steps automatically. If
        such a method is not provided, the design steps must be run manually.

        The desired number of samples to be drawn can either be specified when initializing the
        class instance or when calling this method. If a number of samples is provided on
        both occasions, then the number provided when calling ``run_sequential_design`` is used.

        Internally, this method is a wrapper to ``run_initial_design`` and then calling
        ``run_next_point`` a total of ``n_samples`` times. Note that this means that the total
        number of design points is ``n_init + n_samples``.

        :param n_samples: Number of sequential design steps to be run. Optional if the number was
                          specified upon initialization. Default is ``None`` (default to number
                          set when initializing). If numbers are provided on both occasions, the
                          number set here is used. If a number is provided, must be non-negative.
        :type n_samples: int or None
        :returns: None
        :rtype: None
        """

        assert self.has_function(), "Design must have a bound function to use run_sequential_design"

        if n_samples is None and self.n_samples is None:
            raise ValueError("must specify n_samples either when initializing or calling run_sequential_design")

        if n_samples is None:
            n_iter = self.n_samples
        else:
            n_iter = n_samples

        assert n_iter >= 0, "number of samples must be non-negative"

        self.run_initial_design()

        for i in range(n_iter):
            self.run_next_point()

    def __str__(self):
        """
        Returns string representation of a sequential design

        Returns a string representation of the sequential design. Contains information on the base
        design, the number of points used in the different steps, and the input and target values.

        :returns: String representation of the sequential design
        :rtype: str
        """

        output_string = ""
        output_string += type(self).__name__+" with\n"
        output_string += self.get_base_design()+" base design\n"
        if self.has_function():
            output_string += "a bound simulator function\n"
        output_string += str(self.get_n_samples())+" total samples\n"
        output_string += str(self.get_n_init())+" initial points\n"
        output_string += str(self.get_n_cand())+" candidate points\n"
        output_string += str(self.get_current_iteration())+" current samples\n"
        output_string += "current inputs: "+str(self.get_inputs())+"\n"
        output_string += "current targets: "+str(self.get_targets())

        return output_string


class MICEFastGP(GaussianProcess):
    """
    Derived GaussianProcess class implementing the Woodbury matrix identity for fast predictions

    This class implements a Gaussian Process that is used in the MICE Sequential Design. The GP
    is fit using all candidate points from the sequential design, and the uses the Woodbury
    matrix identity to correct that fit to exclude the candidate point in question. This reduces
    the cost of fitting the GP from O(n^3) to O(n^2), which can dramatically speed up this
    process for large numbers of candidate points. This is mostly used for the particular
    application to the MICE sequential design, but could potentially have other applications
    where many candidate points are to be considered one at a time.
    """
    def fast_predict(self, index):
        """
        Make a fast prediction using one input point to a fit GP

        This method is used to correct a Gaussian Process fit to a set of candidate points to
        evaluate the uncertainty at the candidate point. It is used in the MICE sequential
        design procedure to examine the mutual information between candidate points by determining
        how well correlated the design point is in question to the remainder of the candidates.
        It uses the Woodbury matrix identity to correct the existing GP fit (which requires
        O(n^3) operations) using O(n^2) operations, speeding up the process significantly for
        large candidate design sizes.

        The method requires a fit GP, and the index of the input point that is to be excluded.
        The method then corrects the GP fit and computes the uncertainty of the prediction
        on the excluded point returning the uncertainty as a float.

        :param index: Index of input point to be excluded in the fit and to which the prediction
                      will be applied. Must be an integer with 0 <= index < n (where n is the number
                      of target points in the fit GP, or the number of candidate points when
                      applied to the MICE procedure).
        :type index: int
        :returns: Uncertainty in the corrected fit applied to the given index point
        :rtype: float
        """

        index = int(index)
        assert index >= 0 and index < self.n, "index must be 0 <= index < n"

        indices = (np.arange(self.n) != index)

        switch = self.mean.get_n_params(self.inputs)
        sigma_2 = np.exp(self.theta[-2])

        Ktest = self.kernel.kernel_f(np.reshape(self.inputs[indices,:], (self.n - 1, self.D)),
                                     np.reshape(self.inputs[index, :], (1, self.D)),
                                     self.theta[switch:-1])

        invQ = np.linalg.solve(self.L.T, np.linalg.solve(self.L, np.eye(self.n)))
        invQ_mod = (invQ[indices][:, indices] -
                    1./invQ[index, index]*np.outer(invQ[indices, index], invQ[indices, index]))

        var = np.maximum(sigma_2 - np.sum(Ktest * np.dot(invQ_mod, Ktest), axis=0), 0.)

        return var

class MICEDesign(SequentialDesign):
    """
    Class representing a Mutual Information for Computer Experiments (MICE) sequential
    experimental design

    This class provides an implementation of the MICE algorithm, which uses Mutual Information
    as the criterion for selecting new points in a sequential design. The idea in MICE is to
    select design points based on the point that provides the most information on the function
    values in the entire design space. This is a straightforward application of a sequential
    design procedure, though the class requires a few additional parameters in order to
    compute the MICE criteria.

    These additional parameters are nugget parameters provided to the Gaussian Process fit to
    smooth the predictions when evaluating the Mutual Information criteria. Essentially, since
    experimental design often requires sampling from a high dimensional space, this cannot be
    done in a way that guarantees that all candidate points are equally spaced. The Mutual
    Information criterion is sensitive to how these candidate points are distributed in space,
    so the nugget parameter provides some smoothing that makes the criterion less dependent on
    the distribution of the candidate points. Typical values of the smoothing nugget parameters
    (``nugget_s`` in this implementation) are 1, though this may depend on the application.

    Other than the smoothing parameters, the implementation follows the base procedure for a
    sequential design. The implementation adds methods for querying the nugget parameters
    and an additional helper function for computing the Mutual Information criterion, but
    other methods are identical.
    """
    def __init__(self, base_design, f = None, n_samples = None, n_init = 10, n_cand = 50,
                 nugget = "adaptive", nugget_s = 1.):
        """
        Create new instance of a MICE sequential design

        Method to initialize a new MICE design. Parameters are largely the same as for the base
        ``SequentialDesign`` class, with a few additional nugget parameters for computing the
        Mutual Information criterion. A base design must be provided (must be a subclass of the
        ``ExperimentalDesign`` class), plus optionally a function to be evaluated in the design.
        Additional parameters include the number of samples, the number of initial design points,
        the number of candidate points, the nugget parameter for the base GP, and the smoothing
        nugget parameter for smoothing the uncertainty predictions on the candidate design points.
        Note that the total number of design points is ``n_init + n_samples``.

        :param base_design: Base one-shot experimental design (must be a subclass of
                            ``ExperimentalDesign``). This contains the information on the
                            parameter space to be sampled.
        :type base_design: ExperimentalDesign
        :param f: Function to be evaluated for the design. Must take all parameter values as a single
                  input array and return a single float or an array of length 1
        :type f: function or other callable
        :param n_samples: Number of sequential design points to be drawn. If specified, this must be
                          a positive integer. Note that this is in addition to the number of initial
                          points, meaning that the total design size will be ``n_samples + n_init``.
                          This can also be specified when running the full design. This parameter is
                          optional, and defaults to ``None`` (meaning the number of samples is set when
                          running the design, or that samples will be added manually).
        :type n_samples: int or None
        :param n_init: Number of points in the inital design before the sequential steps begin. Must
                       be a positive integer. Optional, default value is 10.
        :type n_init: int
        :param n_cand: Number of candidates to consider at each sequential design step. Must be a positive
                       integer. Optional, default value is 50.
        :param nugget: Nugget parameter for base GP predictions. Must be a non-negative float or ``None``,
                       where ``None`` indicates that the nugget parameter is selected adaptively. Optional,
                       default value is ``None``.
        :type nugget: float or None
        :param nugget_s: Smoothing nugget parameter for smoothing the predictions on the candidate space.
                         Must be a non-negative float. Default value is 1.
        :type nugget_s: float
        """

        if not isinstance(nugget, str):
            try:
                float(nugget)
            except TypeError:
                raise TypeError("nugget must be a string or convertible to a float")
            if nugget < 0.:
                raise ValueError("nugget parameter cannot be negative")

        if nugget_s < 0.:
            raise ValueError("nugget smoothing parameter cannot be negative")

        if isinstance(nugget, str):
            self.nugget = nugget
        else:
            self.nugget = float(nugget)
        self.nugget_s = float(nugget_s)

        super().__init__(base_design, f, n_samples, n_init, n_cand)

    def get_nugget(self):
        """
        Get value of nugget parameter for base GP

        Returns the nugget value for the base GP (used to actually fit the inputs to targets).
        Can be a float or None (meaning fitting will adaptively add noise to stabilize matrix
        inversion as needed).

        :returns: Nugget parameter, can be a float or None for adaptive noise addition.
        :rtype: float or None
        """
        return self.nugget

    def get_nugget_s(self):
        """
        Get value of smoothing nugget parameter

        Returns the value of the smoothing nugget parameter for the GP used to evaluate the mutual
        information criterion. This GP examines the correlation between a candidate design point and
        the other candidate points, which requires smoothing to ensure that the correlation measure is
        not biased by the distribution of the candidate points in space. This parameter must be a
        nonnegative float (typical values used are 1, though this may depend on the application).

        :returns: Nugget parameter for smoothing predictions from candidate points made on a candidate
                  point. Typical values are 1.
        :rtype: float
        """
        return self.nugget_s

    def _estimate_next_target(self, next_point):
        """
        Estimate value of simulator for a point in a MICE design

        This method is used for the batch version of a sequential design. Instead of updating
        the targets with the known solution, this method is used to estimate the function
        instead. For the MICEDesign, this is just the prediction of the current design GP
        for the point. Returns an array of length 1 holding the prediction.

        :param next_point: Input to be simulated. Must be an array of shape ``(n_parameters,)``
        :type next_point: ndarray
        :returns: Estimated simulation value for the given input as an array of length 1
        :rtype: ndarray
        """

        next_point = np.array(next_point)
        assert next_point.shape == (self.get_n_parameters(),), "bad shape for next_point"

        return self.gp.predict(next_point)[0]

    def _MICE_criterion(self, data_point):
        """
        Compute the MICE criterion for a single candidate point

        This internal method computes the MICE criterion for a single candidate point. Requires
        input of the index of a candidate point to be considered (must be an integer satisfying
        ``0 <= index < n_cand``). It involves fitting a corrected GP to the candidate points other
        than the one under consideration (using the ``MICEFastGP`` class to correct the fit
        via the Woodbury matrix identity), and then computing the MICE criterion based on
        the predictions of the base GP on the point and the corrected GP fit. The MICE
        criterion is then the variance of the base GP divided by the variance of the corrected
        candidate GP. Value returned is the MICE criterion for the point in question.

        :param data_point: Index of the candidate point under consideration. Must be an integer
                           with ``0 <= index < n_cand``.
        :type data_point: int
        :returns: MICE criterion for the data point in question
        :rtype: float
        """

        data_point = int(data_point)

        assert data_point >= 0 and data_point < self.n_cand, "test point index is out of range"

        _, unc1, _ = self.gp.predict(self.candidates[data_point], unc=True, deriv=False)
        unc2 = self.gp_fast.fast_predict(data_point)

        mice_criter =  unc1/unc2

        assert np.isfinite(mice_criter), "error in computing MICE critera"

        return float(mice_criter)

    def _eval_metric(self):
        """
        Evaluate MICE criterion on all candidate points and select new design point

        This internal method computes the MICE criterion on all candidate points and returns
        the index of the point with the maximum value. It does so by first fitting a base GP
        to all points in the current design, and then fitting a dummy GP to all candidate
        design points using the parameter values determined from the base GP fit. The MICE
        criterion does not depend on the target values, since the parameters are determined
        via the base GP and the MICE criterion only depends on the uncertainty of the
        candidate GP (which is independent of the target values). These fit GPs are then used
        to compute the MICE criterion for each candidate point, and the method returns the
        index of the point that had the maximum value of the MICE criterion.

        :returns: Index of the candidate with the maximum MICE score (integer with
                  ``0 <= index < n_cand``)
        :rtype: int
        """

        numtries = 10

        for i in range(numtries):
            try:
                self.gp = GaussianProcess(self.inputs, self.targets, nugget=self.nugget)
                self.gp = fit_GP_MAP(self.gp)

                self.gp_fast = MICEFastGP(self.candidates, np.ones(self.n_cand), nugget=np.exp(self.gp.theta[-2])*self.nugget_s)
                self.gp_fast.theta = self.gp.theta
                break
            except FloatingPointError:
                if i < numtries - 1:
                    continue
                else:
                    raise FloatingPointError("Unable to find parameters suitable for both GPs")
            except LinAlgError:
                if i < numtries - 1:
                    continue
                else:
                    raise LinAlgError("Unable to find parameters suitable for both GPs")

        results = []

        for point in range(self.n_cand):
            results.append(self._MICE_criterion(point))

        return np.argmax(results)



import numpy as np
import pytest
from scipy import linalg
from .. import gKDR
from ..DimensionReduction import median_dist, gram_matrix_sqexp, gram_matrix
from .. import fitting

##### Some simple functions useful for training
def fn(x):
    """A linear function for testing the dimension reduction"""
    return 10*(x[0] + x[1]) + (x[1] - x[0])

def fn2(x):
    """A linear function for testing the dimension reduction"""
    return 10*(x[0] + x[1]) + (x[1] - x[0]) + x[2] + 0.1*x[3]

def fn3(x):
    return x[0]


##### The tests

def test_DimensionReduction_basic():
    """Basic check that we can create gKDR with the expected arguments"""
    Y = np.array([[1],[2.1],[3.2]])
    X = np.array([[1,2,3],[4,5.1,6],[7.1,8,9.1]])
    K_in = 2
    dr = gKDR(X, Y, K=K_in, SGX=2, SGY=2, EPS=1E-5)
    assert(dr.K == K_in)


def test_DimensionReduction_tune_parameters():
    """Check that we can tune the dimension reduction to discover the
    correct structural dimension in a simple case"""

    np.random.seed(100)
    X = np.random.random((20,20))
    Y = np.apply_along_axis(fn3, 1, X)

    dr, loss = gKDR.tune_parameters(X, Y, fitting.fit_GP_MAP,
                                    cXs=[5.0], cYs=[5.0], maxK = 3)

    # These are somewhat conservative bounds: the random seed makes
    # the resulting K and loss deterministic.  For other seeds, these
    # bounds are unlikely to be exceeded.
    assert(dr.K <= 2)
    assert(loss > 0.0 and loss < 0.2)


def test_DimensionReduction_GP():
    """Test that a GP based on reduced inputs behaves well."""

    # Make some test points on a grid.  Any deterministic set of
    # points would work well for this test.

    X = np.mgrid[0:10,0:10].T.reshape(-1,2)/10.0
    Y = np.apply_along_axis(fn, 1, X)
    dr = gKDR(X,Y,1)

    np.random.seed(10)
    gp = fitting.fit_GP_MAP(X, Y)

    gp_red = fitting.fit_GP_MAP(dr(X), Y)

    ## some points offset w.r.t the initial grid
    Xnew = (np.mgrid[0:9,0:9].T.reshape(-1,2) + 0.5)/10.0

    Yexpect = np.apply_along_axis(fn, 1, Xnew)
    Ynew = gp.predict(Xnew)[0] # value prediction
    Ynew_red = gp_red.predict(dr(Xnew))[0] # value prediction

    # check that the fit was reasonable in both cases
    assert(np.max(np.abs(Ynew - Yexpect)) <= 0.02)
    assert(np.max(np.abs(Ynew_red - Yexpect)) <= 0.02)


def test_DimensionReduction_B():
    """Test that a dimension reduction gives the same result as a
    pre-computed result from the Fukumizu matlab code (see
    [Fukumizu1]_)."""

    X = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]])
    Y = np.array([0.1, 1.0, 3.0, 3.6])
    dr = gKDR(X,Y,2,SGX=1.0,SGY=2.0)

    B_expected = np.array([[-0.2653073259794961, -0.9641638982982144],
                           [-0.9641638982982144,  0.2653073259794961]])

    for i in range(B_expected.shape[1]):
        r = dr.B[:,i]/B_expected[:,i]
        assert(np.allclose(r, 1.0) or np.allclose(r, -1.0))

def test_DimensionReduction_median_dist():
    X1 = np.array([[0.0], [1.0], [2.0]])
    assert(np.allclose(median_dist(X1), 1))

    X2 = np.array([[0.0], [1.0], [2.0], [3.0]])
    assert(np.allclose(median_dist(X2), 1.5))

def test_DimensionReduction_gram_matrix():
    X = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]])
    def k_dot(x0, x1):
        return np.dot(x0, x1)

    def k_sqexp(x0, x1):
        d = x0 - x1
        return np.exp(-0.5 * np.dot(d, d))

    G_dot = gram_matrix(X, k_dot)
    assert(np.allclose(G_dot, np.array([[0.0, 0.0, 0.0, 0.0],
                                        [0.0, 1.0, 0.0, 1.0],
                                        [0.0, 0.0, 1.0, 1.0],
                                        [0.0, 1.0, 1.0, 2.0]])))

    G_sqexp1 = gram_matrix_sqexp(X, 1.0)
    G_sqexp2 = gram_matrix(X, k_sqexp)
    G_sqexp_expected = np.exp(np.array([[ 0.0, -0.5, -0.5, -1.0],
                                        [-0.5,  0.0, -1.0, -0.5],
                                        [-0.5, -1.0,  0.0, -0.5],
                                        [-1.0, -0.5, -0.5,  0.0]]))

    assert(np.allclose(G_sqexp1, G_sqexp_expected))
    assert(np.allclose(G_sqexp2, G_sqexp_expected))



import numpy as np
import pytest
from numpy.testing import assert_allclose
from ..GaussianProcess import GaussianProcess, PredictResult
from ..MultiOutputGP import MultiOutputGP
from ..MeanFunction import ConstantMean, LinearMean, MeanFunction
from ..Kernel import Matern52
from scipy import linalg

@pytest.fixture
def x():
    return np.array([[1., 2., 3.], [4., 5., 6.]])

@pytest.fixture
def y():
    return np.array([[2., 4.], [3., 5.]])

def test_MultiOutputGP_init(x, y):
    "Test function for correct functioning of the init method of GaussianProcess"

    gp = MultiOutputGP(x, y)
    assert_allclose(x, gp.emulators[0].inputs)
    assert_allclose(x, gp.emulators[1].inputs)
    assert_allclose(y[0], gp.emulators[0].targets)
    assert_allclose(y[1], gp.emulators[1].targets)
    assert gp.D == 3
    assert gp.n == 2
    assert gp.n_emulators == 2

@pytest.fixture
def dx():
    return 1.e-6

def test_MultiOutputGP_predict(x, y, dx):
    "test the predict method of GaussianProcess"

    gp = MultiOutputGP(x, y, nugget=0.)
    theta = np.ones(gp.emulators[0].n_params)

    gp.emulators[0].fit(theta)
    gp.emulators[1].fit(theta)

    x_test = np.array([[2., 3., 4.]])

    mu, var, deriv = gp.predict(x_test)

    for i in range(2):

        K = gp.emulators[i].kernel.kernel_f(x, x, theta[:-1])
        Ktest = gp.emulators[i].kernel.kernel_f(x_test, x, theta[:-1])

        mu_expect = np.dot(Ktest, gp.emulators[i].invQt)
        var_expect = np.exp(theta[-2]) - np.diag(np.dot(Ktest, np.linalg.solve(K, Ktest.T)))

        D = gp.D

        deriv_expect = np.zeros((1, D))

        for j in range(D):
            dx_array = np.zeros(D)
            dx_array[j] = dx
            deriv_expect[0, j] = (gp.emulators[i].predict(x_test)[0] - gp.emulators[i].predict(x_test - dx_array)[0])/dx

        assert_allclose(mu[i], mu_expect)
        assert_allclose(var[i], var_expect)
        assert_allclose(deriv[i], deriv_expect, atol=1.e-6, rtol=1.e-6)

    nugget = 1.
    gp = MultiOutputGP(x, y, nugget=nugget)
    theta = np.ones(gp.emulators[0].n_params)

    gp.emulators[0].fit(theta)
    gp.emulators[1].fit(theta)

    x_test = np.array([[2., 3., 4.]])

    mu, var, deriv = gp.predict(x_test)

    for i in range(2):

        K = gp.emulators[i].kernel.kernel_f(x, x, theta[:-1]) + np.eye(gp.emulators[i].n)*nugget
        Ktest = gp.emulators[i].kernel.kernel_f(x_test, x, theta[:-1])

        var_expect = np.exp(theta[-2]) + nugget - np.diag(np.dot(Ktest, np.linalg.solve(K, Ktest.T)))

        assert_allclose(var[i], var_expect)

    mu, var, deriv = gp.predict(x_test, include_nugget=False)

    for i in range(2):

        K = gp.emulators[i].kernel.kernel_f(x, x, theta[:-1]) + np.eye(gp.emulators[i].n)*nugget
        Ktest = gp.emulators[i].kernel.kernel_f(x_test, x, theta[:-1])

        var_expect = np.exp(theta[-2]) - np.diag(np.dot(Ktest, np.linalg.solve(K, Ktest.T)))

        assert_allclose(var[i], var_expect)



import setuptools

# version information
MAJOR = 0
MINOR = 4
MICRO = 0
PRERELEASE = 0
ISRELEASED = True
version = "{}.{}.{}".format(MAJOR, MINOR, MICRO)

if not ISRELEASED:
    version += ".dev{}".format(PRERELEASE)

# write version information to file

def write_version_file(version):
    "writes version file that is read when importing version number"
    version_file = """'''
Version file automatically created by setup.py file
'''
version = '{}'
    """.format(version)

    with open("mogp_emulator/version.py", "w") as fh:
        fh.write(version_file)

write_version_file(version)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(name='mogp_emulator',
      version=version,
      description='Tool for Multi-Output Gaussian Process Emulators',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://mogp-emulator.readthedocs.io/',
      project_urls={
                  "Bug Tracker": "https://github.com/alan-turing-institute/mogp-emulator/issues",
                  "Documentation": "https://mogp-emulator.readthedocs.io/",
                  "Source Code": "https://github.com/alan-turing-institute/mogp-emulator/",
              },
      author='Alan Turing Institute Research Engineering Group',
      author_email='edaub@turing.ac.uk',
      packages=setuptools.find_packages(),
      license=['MIT'],
      install_requires=['numpy', 'scipy'])
